#! /usr/bin/env bash
# This file contains the complete sequence of commands
# Mentor Graphics used to build this version of Sourcery CodeBench.
# 
# For each free or open-source component of Sourcery CodeBench,
# the source code provided includes all of the configuration
# scripts and makefiles for that component, including any and
# all modifications made by Mentor Graphics.  From this list of
# commands, you can see every configuration option used by
# Mentor Graphics during the build process.
# 
# This file is provided as a guideline for users who wish to
# modify and rebuild a free or open-source component of
# Sourcery CodeBench from source. For a number of reasons,
# though, you may not be able to successfully run this script
# directly on your system. Certain aspects of the Mentor Graphics
# build environment (such as directory names) are included in
# these commands. Mentor Graphics uses Canadian cross compilers so
# you may need to modify various configuration options and paths
# if you are building natively. This sequence of commands
# includes those used to build proprietary components of
# Sourcery CodeBench for which source code is not provided.
# 
# Please note that Sourcery CodeBench support covers only your
# use of the original, validated binaries provided as part of
# Sourcery CodeBench -- and specifically does not cover either
# the process of rebuilding a component or the use of any
# binaries you may build.  In addition, if you rebuild any
# component, you must not use the --with-pkgversion and
# --with-bugurl configuration options that embed Mentor Graphics
# trademarks in the resulting binary; see the "Mentor Graphics
# Trademarks" section in the Sourcery CodeBench Software
# License Agreement.
set -e
inform_fd=2 
umask 022
exec < /dev/null

error_handler ()
{
    exit 1
}

check_status() {
    local status="$?"
    if [ "$status" -ne 0 ]; then
	error_handler
    fi
}

check_pipe() {
    local -a status=("${PIPESTATUS[@]}")
    local limit=$1
    local ix
    
    if [ -z "$limit" ] ; then
	limit="${#status[@]}"
    fi
    for ((ix=0; ix != $limit ; ix++)); do
	if [ "${status[$ix]}" != "0" ] ; then
	    error_handler
	fi
    done
}

error () {
    echo "$script: error: $@" >& $inform_fd
    exit 1
}

nonfatal_error () {
    echo "$script: error: $@" >& $inform_fd
}

warning () {
    echo "$script: warning: $@" >& $inform_fd
}

verbose () {
    if $gnu_verbose; then
	echo "$script: $@" >& $inform_fd
    fi
}

copy_dir() {
    mkdir -p "$2"

    (cd "$1" && tar cf - .) | (cd "$2" && tar xf -)
    check_pipe
}

copy_dir_clean() {
    mkdir -p "$2"
    (cd "$1" && tar cf - \
	--exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc \
	--exclude="*~" --exclude=".#*" \
	--exclude="*.orig" --exclude="*.rej" \
	.) | (cd "$2" && tar xf -)
    check_pipe
}

update_dir_clean() {
    mkdir -p "$2"


    (cd "$1" && tar cf - \
	--exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc \
	--exclude="*~" --exclude=".#*" \
	--exclude="*.orig" --exclude="*.rej" \
	--after-date="$3" \
	. 2> /dev/null) | (cd "$2" && tar xf -)
    check_pipe
}

copy_dir_exclude() {
    local source="$1"
    local dest="$2"
    local excl="$3"
    shift 3
    mkdir -p "$dest"
    (cd "$source" && tar cfX - "$excl" "$@") | (cd "$dest" && tar xf -)
    check_pipe
}

copy_dir_only() {
    local source="$1"
    local dest="$2"
    shift 2
    mkdir -p "$dest"
    (cd "$source" && tar cf - "$@") | (cd "$dest" && tar xf -)
    check_pipe
}

check_mentor_trademarks ()
{
    local cfg_script="$1/configure"
    shift
    local have_mentor_trademark=false
    local c
    for c in "$@"; do
	case "$c" in
	    (--with-pkgversion=*[Ss]ourcery*)
		have_mentor_trademark=true
		;;
	    (--with-bugurl=*[Ss]ourcery*)
		have_mentor_trademark=true
		;;
	    (*)
		;;
	esac
    done
    if "$have_mentor_trademark"; then
	case $(dnsdomainname) in
	    (*.mentorg.com)
		"$cfg_script" "$@"
		;;
	    (*)
		error "If you rebuild any component, you must not use the" \
		    "--with-pkgversion and --with-bugurl configuration" \
		    "options that embed Mentor Graphics trademarks in the" \
		    "resulting binary; see the \"Mentor Graphics" \
		    "Trademarks\" section in the Software License Agreement."
		;;
	esac
    else
	"$cfg_script" "$@"
    fi
}

clean_environment() {
    local env_var_list
    local var




    unset BASH_ENV CDPATH POSIXLY_CORRECT TMOUT

    env_var_list=$(export | \
	grep '^declare -x ' | \
	sed -e 's/^declare -x //' -e 's/=.*//')

    for var in $env_var_list; do
	case $var in
	    HOME|HOSTNAME|LOGNAME|PWD|SHELL|SHLVL|SSH_*|TERM|USER)


		;;
	    LD_LIBRARY_PATH|PATH| \
		FLEXLM_NO_CKOUT_INSTALL_LIC|LM_APP_DISABLE_CACHE_READ)


		;;
	    CSL_CONFIG_SUB)

		;;
	    MAKEINFO)

		;;
	    *_LICENSE_FILE)













		if false; then
		    local license_file_envvar
		    license_file_envvar=

		    if [ "$var" != "$license_file_envvar" ]; then
			export -n "$var" || true
		    fi
		else
		    export -n "$var" || true
		fi
		;;
	    *)

		export -n "$var" || true
		;;
	esac
    done


    export LANG=C
    export LC_ALL=C


    export CVS_RSH=ssh



    user_shell=$SHELL
    export SHELL=$BASH
    export CONFIG_SHELL=$BASH
}

pushenv() {
    pushenv_level=$(($pushenv_level + 1))
    eval pushenv_vars_${pushenv_level}=
}


pushenv_level=0
pushenv_vars_0=



pushenvvar() {
    local pushenv_var="$1"
    local pushenv_newval="$2"
    eval local pushenv_oldval=\"\$$pushenv_var\"
    eval local pushenv_oldset=\"\${$pushenv_var+set}\"
    local pushenv_save_var=saved_${pushenv_level}_${pushenv_var}
    local pushenv_savep_var=savedp_${pushenv_level}_${pushenv_var}
    eval local pushenv_save_set=\"\${$pushenv_savep_var+set}\"
    if [ "$pushenv_save_set" = "set" ]; then
	error "Pushing $pushenv_var more than once at level $pushenv_level"
    fi
    if [ "$pushenv_oldset" = "set" ]; then
	eval $pushenv_save_var=\"\$pushenv_oldval\"
    else
	unset $pushenv_save_var
    fi
    eval $pushenv_savep_var=1
    eval export $pushenv_var=\"\$pushenv_newval\"
    local pushenv_list_var=pushenv_vars_${pushenv_level}
    eval $pushenv_list_var=\"\$$pushenv_list_var \$pushenv_var\"
}

prependenvvar() {
    local pushenv_var="$1"
    local pushenv_newval="$2"
    eval local pushenv_oldval=\"\$$pushenv_var\"
    pushenvvar "$pushenv_var" "$pushenv_newval$pushenv_oldval"
}

popenv() {
    local pushenv_var=
    eval local pushenv_vars=\"\$pushenv_vars_${pushenv_level}\"
    for pushenv_var in $pushenv_vars; do
	local pushenv_save_var=saved_${pushenv_level}_${pushenv_var}
	local pushenv_savep_var=savedp_${pushenv_level}_${pushenv_var}
	eval local pushenv_save_val=\"\$$pushenv_save_var\"
	eval local pushenv_save_set=\"\${$pushenv_save_var+set}\"
	unset $pushenv_save_var
	unset $pushenv_savep_var
	if [ "$pushenv_save_set" = "set" ]; then
	    eval export $pushenv_var=\"\$pushenv_save_val\"
	else
	    unset $pushenv_var
	fi
    done
    unset pushenv_vars_${pushenv_level}
    if [ "$pushenv_level" = "0" ]; then
	error "Popping environment level 0"
    else
	pushenv_level=$(($pushenv_level - 1))
    fi
}

prepend_path() {
    if $(eval "test -n \"\$$1\""); then
	prependenvvar "$1" "$2:"
    else
	prependenvvar "$1" "$2"
    fi
}
pushenvvar CSL_SCRIPTDIR /scratch/sandra/nios2-linux-bugfix/src/scripts-trunk
pushenvvar PATH /usr/local/tools/gcc-4.7.3/bin
pushenvvar LD_LIBRARY_PATH /usr/local/tools/gcc-4.7.3/i686-pc-linux-gnu/lib:/usr/local/tools/gcc-4.7.3/lib64:/usr/local/tools/gcc-4.7.3/lib
pushenvvar FLEXLM_NO_CKOUT_INSTALL_LIC 1
pushenvvar LM_APP_DISABLE_CACHE_READ 1
pushenvvar MAKEINFO 'makeinfo --css-ref=../cs.css'
pushenvvar CSL_CONFIG_SUB /scratch/sandra/nios2-linux-bugfix/src/config-trunk/config.sub
clean_environment
# task [001/219] /init/dirs
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj
mkdir -p /scratch/sandra/nios2-linux-bugfix/install
mkdir -p /scratch/sandra/nios2-linux-bugfix/pkg
mkdir -p /scratch/sandra/nios2-linux-bugfix/logs/data
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf
popenv

# task [002/219] /init/cleanup
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/pkg/sourceryg++-2013.05-43-nios2-linux-gnu.src.tar.bz2 /scratch/sandra/nios2-linux-bugfix/pkg/sourceryg++-2013.05-43-nios2-linux-gnu.backup.tar.bz2
popenv

# task [003/219] /init/source_package/binutils
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/binutils-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/binutils-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' binutils-2013.05
popd
popenv

# task [004/219] /init/source_package/gcc
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/gcc-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/gcc-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' gcc-nios2
popd
popenv

# task [005/219] /init/source_package/gdb
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/gdb-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/gdb-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' gdb-nios2
popd
popenv

# task [006/219] /init/source_package/zlib
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/zlib-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/zlib-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' zlib-1.2.7
popd
popenv

# task [007/219] /init/source_package/gmp
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/gmp-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/gmp-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' gmp-2013.05
popd
popenv

# task [008/219] /init/source_package/mpfr
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/mpfr-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/mpfr-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' mpfr-2013.05
popd
popenv

# task [009/219] /init/source_package/mpc
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/mpc-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/mpc-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' mpc-2013.05
popd
popenv

# task [010/219] /init/source_package/cloog
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/cloog-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/cloog-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' cloog-0.15
popd
popenv

# task [011/219] /init/source_package/ppl
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/ppl-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/ppl-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' ppl-sgxx-0.11-2013.05
popd
popenv

# task [012/219] /init/source_package/ncurses
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/ncurses-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/ncurses-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' ncurses-5.9
popd
popenv

# task [013/219] /init/source_package/getting_started
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/getting_started-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/getting_started-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' getting-started-2013.05
popd
popenv

# task [014/219] /init/source_package/installanywhere
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/installanywhere-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/installanywhere-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' installanywhere-trunk
popd
popenv

# task [015/219] /init/source_package/cs3
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/cs3-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/cs3-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' cs3-2013.05
popd
popenv

# task [016/219] /init/source_package/sprite
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/sprite-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/sprite-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' sprite-2013.05
popd
popenv

# task [017/219] /init/source_package/glibc
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/glibc-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/glibc-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' glibc-trunk-nios2-2_17
popd
popenv

# task [018/219] /init/source_package/glibc_localedef
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/glibc_localedef-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/glibc_localedef-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' glibc_localedef-trunk-nios2-2_17
popd
popenv

# task [019/219] /init/source_package/linux
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/linux-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/linux-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' linux-nios2
popd
popenv

# task [020/219] /init/source_package/coremark
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/coremark-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/coremark-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' coremark-trunk
popd
popenv

# task [021/219] /init/source_package/csl_tests
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/csl_tests-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/csl_tests-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' csl-tests-trunk
popd
popenv

# task [022/219] /init/source_package/dejagnu_boards
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/dejagnu_boards-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/dejagnu_boards-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' dejagnu-boards-trunk
popd
popenv

# task [023/219] /init/source_package/scripts
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/scripts-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/scripts-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' scripts-trunk
popd
popenv

# task [024/219] /init/source_package/xfails
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/xfails-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/xfails-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' xfails-trunk
popd
popenv

# task [025/219] /init/source_package/portal
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/portal-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/portal-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' portal-trunk
popd
popenv

# task [026/219] /init/source_package/config
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/config-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/config-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' config-trunk
popd
popenv

# task [027/219] /init/source_package/libiconv
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/libiconv-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/libiconv-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' libiconv-1.11
popd
popenv

# task [028/219] /init/source_package/libelf
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/libelf-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/libelf-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' libelf-2013.05
popd
popenv

# task [029/219] /init/source_package/expat
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/expat-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/expat-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' expat-2013.05
popd
popenv

# task [030/219] /init/source_package/csl_docbook
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/csl_docbook-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/csl_docbook-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' csl-docbook-trunk
popd
popenv

# task [031/219] /init/source_package/release_notes
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/release_notes-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/release_notes-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' release-notes-trunk
popd
popenv

# task [032/219] /init/source_package/make
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/make-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/make-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' make-3.81
popd
popenv

# task [033/219] /init/source_package/coreutils
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/coreutils-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu/coreutils-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' coreutils-5.94
popd
popenv

# task [034/219] /init/source_package/gdil
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/gdil-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/gdil-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' gdil-2013.05
popd
popenv

# task [035/219] /init/source_package/edge_utils
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/edge_utils-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/edge_utils-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' edge-utils-2013.05
popd
popenv

# task [036/219] /init/source_package/eglibc_configs
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
rm -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/eglibc_configs-2013.05-43.tar.bz2
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup
pushd /scratch/sandra/nios2-linux-bugfix/src
tar cf /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup/eglibc_configs-2013.05-43.tar.bz2 --bzip2 --owner=0 --group=0 --exclude=CVS --exclude=.svn --exclude=.git --exclude=.pc '--exclude=*~' '--exclude=.#*' '--exclude=*.orig' '--exclude=*.rej' eglibc-configs-trunk
popd
popenv

# task [037/219] /i686-pc-linux-gnu/host_cleanup
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
popenv

# task [038/219] /i686-pc-linux-gnu/zlib_first/copy
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/zlib-first-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
copy_dir_clean /scratch/sandra/nios2-linux-bugfix/src/zlib-1.2.7 /scratch/sandra/nios2-linux-bugfix/obj/zlib-first-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
chmod -R u+w /scratch/sandra/nios2-linux-bugfix/obj/zlib-first-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
popenv

# task [039/219] /i686-pc-linux-gnu/zlib_first/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushd /scratch/sandra/nios2-linux-bugfix/obj/zlib-first-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushenv
pushenvvar CFLAGS '-O3 -fPIC'
pushenvvar CC 'i686-pc-linux-gnu-gcc  '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
./configure --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --static
popenv
popd
popenv

# task [040/219] /i686-pc-linux-gnu/zlib_first/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushd /scratch/sandra/nios2-linux-bugfix/obj/zlib-first-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4
popd
popenv

# task [041/219] /i686-pc-linux-gnu/zlib_first/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushd /scratch/sandra/nios2-linux-bugfix/obj/zlib-first-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make install
popd
popenv

# task [042/219] /i686-pc-linux-gnu/gmp/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushenvvar CFLAGS '-g -O2'
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/gmp-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/gmp-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/gmp-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/gmp-2013.05 --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --disable-shared --build=i686-pc-linux-gnu --target=i686-pc-linux-gnu --host=i686-pc-linux-gnu --enable-cxx --disable-nls
popd
popenv
popenv
popenv

# task [043/219] /i686-pc-linux-gnu/gmp/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushenvvar CFLAGS '-g -O2'
pushd /scratch/sandra/nios2-linux-bugfix/obj/gmp-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4
popd
popenv
popenv
popenv

# task [044/219] /i686-pc-linux-gnu/gmp/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushenvvar CFLAGS '-g -O2'
pushd /scratch/sandra/nios2-linux-bugfix/obj/gmp-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make install
popd
popenv
popenv
popenv

# task [045/219] /i686-pc-linux-gnu/gmp/postinstall
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushenvvar CFLAGS '-g -O2'
pushd /scratch/sandra/nios2-linux-bugfix/obj/gmp-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make check
popd
popenv
popenv
popenv

# task [046/219] /i686-pc-linux-gnu/mpfr/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/mpfr-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/mpfr-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/mpfr-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/mpfr-2013.05 --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --disable-shared --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-pc-linux-gnu --disable-nls --with-gmp=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr
popd
popenv
popenv
popenv

# task [047/219] /i686-pc-linux-gnu/mpfr/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/mpfr-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4
popd
popenv
popenv
popenv

# task [048/219] /i686-pc-linux-gnu/mpfr/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/mpfr-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make install
popd
popenv
popenv
popenv

# task [049/219] /i686-pc-linux-gnu/mpfr/postinstall
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/mpfr-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make check
popd
popenv
popenv
popenv

# task [050/219] /i686-pc-linux-gnu/mpc/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/mpc-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/mpc-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/mpc-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/mpc-2013.05 --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --disable-shared --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-pc-linux-gnu --disable-nls --with-gmp=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --with-mpfr=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr
popd
popenv
popenv
popenv

# task [051/219] /i686-pc-linux-gnu/mpc/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/mpc-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4
popd
popenv
popenv
popenv

# task [052/219] /i686-pc-linux-gnu/mpc/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/mpc-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make install
popd
popenv
popenv
popenv

# task [053/219] /i686-pc-linux-gnu/mpc/postinstall
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/mpc-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make check
popd
popenv
popenv
popenv

# task [054/219] /i686-pc-linux-gnu/ppl/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/ppl-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/ppl-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/ppl-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/ppl-sgxx-0.11-2013.05 --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --disable-shared --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-pc-linux-gnu --disable-nls --with-libgmp-prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --with-gmp-prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr CPPFLAGS=-I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/include LDFLAGS=-L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/lib --disable-watchdog
popd
popenv
popenv
popenv

# task [055/219] /i686-pc-linux-gnu/ppl/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/ppl-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4
popd
popenv
popenv
popenv

# task [056/219] /i686-pc-linux-gnu/ppl/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/ppl-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make install
popd
popenv
popenv
popenv

# task [057/219] /i686-pc-linux-gnu/cloog/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/cloog-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/cloog-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/cloog-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/cloog-0.15 --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --disable-shared --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-pc-linux-gnu --disable-nls --with-ppl=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --with-gmp=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr
popd
popenv
popenv
popenv

# task [058/219] /i686-pc-linux-gnu/cloog/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/cloog-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4
popd
popenv
popenv
popenv

# task [059/219] /i686-pc-linux-gnu/cloog/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/cloog-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make install
popd
popenv
popenv
popenv

# task [060/219] /i686-pc-linux-gnu/cloog/postinstall
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/cloog-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make check
popd
popenv
popenv
popenv

# task [061/219] /i686-pc-linux-gnu/libelf/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/libelf-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/libelf-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/libelf-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/libelf-2013.05 --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --disable-shared --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-pc-linux-gnu --disable-nls
popd
popenv
popenv
popenv

# task [062/219] /i686-pc-linux-gnu/libelf/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/libelf-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4
popd
popenv
popenv
popenv

# task [063/219] /i686-pc-linux-gnu/libelf/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/libelf-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make install
popd
popenv
popenv
popenv

# task [064/219] /i686-pc-linux-gnu/toolchain/binutils/copy
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/lib
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/binutils-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
copy_dir_clean /scratch/sandra/nios2-linux-bugfix/src/binutils-2013.05 /scratch/sandra/nios2-linux-bugfix/obj/binutils-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
chmod -R u+w /scratch/sandra/nios2-linux-bugfix/obj/binutils-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
touch /scratch/sandra/nios2-linux-bugfix/obj/binutils-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/.gnu-stamp
popenv
popenv
popenv

# task [065/219] /i686-pc-linux-gnu/toolchain/binutils/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/lib
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/obj/binutils-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu --prefix=/opt/codesourcery --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-pc-linux-gnu '--with-pkgversion=Sourcery CodeBench Lite 2013.05-43' --with-bugurl=https://sourcery.mentor.com/GNUToolchain/ --disable-nls --with-sysroot=/opt/codesourcery/nios2-linux-gnu/libc --enable-poison-system-directories --enable-plugins
popd
popenv
popenv
popenv

# task [066/219] /i686-pc-linux-gnu/toolchain/binutils/libiberty
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/lib
pushd /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4 all-libiberty
popd
copy_dir_clean /scratch/sandra/nios2-linux-bugfix/src/binutils-2013.05/include /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-pc-linux-gnu/usr/include
chmod -R u+w /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-pc-linux-gnu/usr/include
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-pc-linux-gnu/usr/lib
cp /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/libiberty/libiberty.a /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-pc-linux-gnu/usr/lib
popenv
popenv
popenv

# task [067/219] /i686-pc-linux-gnu/toolchain/binutils/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/lib
pushd /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4
popd
popenv
popenv
popenv

# task [068/219] /i686-pc-linux-gnu/toolchain/binutils/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/lib
pushd /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make install prefix=/scratch/sandra/nios2-linux-bugfix/install exec_prefix=/scratch/sandra/nios2-linux-bugfix/install libdir=/scratch/sandra/nios2-linux-bugfix/install/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/man datadir=/scratch/sandra/nios2-linux-bugfix/install/share
popd
popenv
popenv
popenv

# task [069/219] /i686-pc-linux-gnu/toolchain/binutils/postinstall
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/lib
pushd /scratch/sandra/nios2-linux-bugfix/install
rm ./lib/libiberty.a
rmdir ./lib
popd
pushd /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make prefix=/scratch/sandra/nios2-linux-bugfix/install exec_prefix=/scratch/sandra/nios2-linux-bugfix/install libdir=/scratch/sandra/nios2-linux-bugfix/install/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/man datadir=/scratch/sandra/nios2-linux-bugfix/install/share install-html
popd
pushd /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make prefix=/scratch/sandra/nios2-linux-bugfix/install exec_prefix=/scratch/sandra/nios2-linux-bugfix/install libdir=/scratch/sandra/nios2-linux-bugfix/install/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/man datadir=/scratch/sandra/nios2-linux-bugfix/install/share install-pdf
popd
cp /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/bfd/.libs/libbfd.a /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-pc-linux-gnu/usr/lib
cp /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/bfd/bfd.h /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-pc-linux-gnu/usr/include
cp /scratch/sandra/nios2-linux-bugfix/src/binutils-2013.05/bfd/elf-bfd.h /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-pc-linux-gnu/usr/include
cp /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/opcodes/.libs/libopcodes.a /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-pc-linux-gnu/usr/lib
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-pc-linux-gnu/usr/testbin
cp /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/binutils/bfdtest1 /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-pc-linux-gnu/usr/testbin
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-pc-linux-gnu/usr/testbin
cp /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/binutils/bfdtest2 /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-pc-linux-gnu/usr/testbin
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/configure.html /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/etc/configure.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info/configure.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly configure
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/standards.html /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/etc/standards.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info/standards.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly standards
rmdir /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/etc
rm -rf /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/bfd.html /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/bfd.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info/bfd.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly bfd
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/libiberty.html /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/libiberty.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-ld.bfd
rm -f /scratch/sandra/nios2-linux-bugfix/install/bin/ld.bfd
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin/ld.bfd
popenv
popenv
popenv

# task [070/219] /i686-pc-linux-gnu/toolchain/install_kernel_headers_glibc
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/include /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/include/bits /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/include/gnu
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/linux-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
copy_dir_clean /scratch/sandra/nios2-linux-bugfix/src/linux-nios2 /scratch/sandra/nios2-linux-bugfix/obj/linux-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
chmod -R u+w /scratch/sandra/nios2-linux-bugfix/obj/linux-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/linux-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make ARCH=nios2 CROSS_COMPILE=/scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu- INSTALL_HDR_PATH=/scratch/sandra/nios2-linux-bugfix/obj/linux-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/tmp-install headers_install
pushd /scratch/sandra/nios2-linux-bugfix/obj/linux-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/tmp-install/include
copy_dir linux /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/include/linux
copy_dir asm /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/include/asm
copy_dir asm-generic /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/include/asm-generic
copy_dir drm /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/include/drm
copy_dir mtd /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/include/mtd
copy_dir rdma /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/include/rdma
copy_dir sound /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/include/sound
copy_dir video /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/include/video
popd
popd
popenv

# task [071/219] /i686-pc-linux-gnu/toolchain/create_sysroot/glibc_0
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./
popenv

# task [072/219] /i686-pc-linux-gnu/toolchain/gcc_first/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenvvar AR_FOR_TARGET nios2-linux-gnu-ar
pushenvvar NM_FOR_TARGET nios2-linux-gnu-nm
pushenvvar OBJDUMP_FOR_TARET nios2-linux-gnu-objdump
pushenvvar STRIP_FOR_TARGET nios2-linux-gnu-strip
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/gcc-first-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/gcc-first-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/gcc-first-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/gcc-nios2 --build=i686-pc-linux-gnu --host=i686-pc-linux-gnu --target=nios2-linux-gnu --enable-threads --disable-libmudflap --disable-libssp --disable-libstdcxx-pch --with-gnu-as --with-gnu-ld '--with-specs=%{save-temps: -fverbose-asm} -D__CS_SOURCERYGXX_MAJ__=2013 -D__CS_SOURCERYGXX_MIN__=5 -D__CS_SOURCERYGXX_REV__=43 %{O2:%{!fno-remove-local-statics: -fremove-local-statics}} %{O*:%{O|O0|O1|O2|Os:;:%{!fno-remove-local-statics: -fremove-local-statics}}}' --enable-languages=c,c++ --enable-shared --enable-lto --enable-symvers=gnu --enable-__cxa_atexit '--with-pkgversion=Sourcery CodeBench Lite 2013.05-43' --with-bugurl=https://sourcery.mentor.com/GNUToolchain/ --disable-nls --prefix=/opt/codesourcery --disable-shared --disable-threads --disable-libssp --disable-libgomp --without-headers --with-newlib --disable-decimal-float --disable-libffi --disable-libquadmath --disable-libitm --disable-libatomic --enable-languages=c --with-sysroot=/opt/codesourcery/nios2-linux-gnu/libc --with-build-sysroot=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc --with-gmp=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --with-mpfr=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --with-mpc=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --with-ppl=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr '--with-host-libstdcxx=-static-libgcc -Wl,-Bstatic,-lstdc++,-Bdynamic -lm' --with-cloog=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --with-libelf=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --disable-libgomp --disable-libitm --enable-poison-system-directories --with-build-time-tools=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin --with-build-time-tools=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin SED=sed
popd
popenv
popenv

# task [073/219] /i686-pc-linux-gnu/toolchain/gcc_first/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenvvar AR_FOR_TARGET nios2-linux-gnu-ar
pushenvvar NM_FOR_TARGET nios2-linux-gnu-nm
pushenvvar OBJDUMP_FOR_TARET nios2-linux-gnu-objdump
pushenvvar STRIP_FOR_TARGET nios2-linux-gnu-strip
pushd /scratch/sandra/nios2-linux-bugfix/obj/gcc-first-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4 LDFLAGS_FOR_TARGET=--sysroot=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc CPPFLAGS_FOR_TARGET=--sysroot=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc build_tooldir=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu
popd
popenv
popenv

# task [074/219] /i686-pc-linux-gnu/toolchain/gcc_first/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenvvar AR_FOR_TARGET nios2-linux-gnu-ar
pushenvvar NM_FOR_TARGET nios2-linux-gnu-nm
pushenvvar OBJDUMP_FOR_TARET nios2-linux-gnu-objdump
pushenvvar STRIP_FOR_TARGET nios2-linux-gnu-strip
pushd /scratch/sandra/nios2-linux-bugfix/obj/gcc-first-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make prefix=/scratch/sandra/nios2-linux-bugfix/install exec_prefix=/scratch/sandra/nios2-linux-bugfix/install libdir=/scratch/sandra/nios2-linux-bugfix/install/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/man install
make prefix=/scratch/sandra/nios2-linux-bugfix/install exec_prefix=/scratch/sandra/nios2-linux-bugfix/install libdir=/scratch/sandra/nios2-linux-bugfix/install/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/man install-html
make prefix=/scratch/sandra/nios2-linux-bugfix/install exec_prefix=/scratch/sandra/nios2-linux-bugfix/install libdir=/scratch/sandra/nios2-linux-bugfix/install/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/man install-pdf
popd
popenv
popenv

# task [075/219] /i686-pc-linux-gnu/toolchain/gcc_first/postinstall
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenvvar AR_FOR_TARGET nios2-linux-gnu-ar
pushenvvar NM_FOR_TARGET nios2-linux-gnu-nm
pushenvvar OBJDUMP_FOR_TARET nios2-linux-gnu-objdump
pushenvvar STRIP_FOR_TARGET nios2-linux-gnu-strip
pushd /scratch/sandra/nios2-linux-bugfix/install
rm ./lib/libiberty.a
rmdir include
popd
rm -rf /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/gccinstall /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/gcc/gccinstall.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info/gccinstall.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly gccinstall
rm -rf /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/gccint /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/gcc/gccint.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info/gccint.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly gccint
rm -rf /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/cppinternals /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/gcc/cppinternals.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info/cppinternals.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly cppinternals
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/libiberty.html /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/libiberty.pdf
popenv
popenv

# task [076/219] /i686-pc-linux-gnu/toolchain/glibc_headers/glibc_0/copy
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/glibc-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/default/
copy_dir_clean /scratch/sandra/nios2-linux-bugfix/src/glibc-trunk-nios2-2_17 /scratch/sandra/nios2-linux-bugfix/obj/glibc-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/default/
chmod -R u+w /scratch/sandra/nios2-linux-bugfix/obj/glibc-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/default/
touch /scratch/sandra/nios2-linux-bugfix/obj/glibc-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/default//.gnu-stamp
popenv

# task [077/219] /i686-pc-linux-gnu/toolchain/glibc_headers/glibc_0/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./lib/./ /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/glibc-headers-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/default/
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/glibc-headers-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/default/
pushd /scratch/sandra/nios2-linux-bugfix/obj/glibc-headers-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/default/
pushenv
pushenvvar BUILD_CC gcc
pushenvvar CC '/scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-gcc '
pushenvvar CXX '/scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-g++ '
pushenvvar CFLAGS '-g -O2'
pushenvvar AR /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-ar
pushenvvar NM /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-nm
pushenvvar RANLIB /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-ranlib
pushenvvar READELF /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-readelf
pushenvvar PERL /usr/bin/perl
pushenvvar BASH_SHELL /bin/bash
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/obj/glibc-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/default/ --prefix=/usr --with-headers=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/include --build=i686-pc-linux-gnu --host=nios2-linux-gnu --disable-profile --without-gd --without-cvs --enable-obsolete-rpc --enable-add-ons --enable-kernel=2.6.21 '--with-pkgversion=Sourcery CodeBench Lite 2013.05-43' --with-bugurl=https://sourcery.mentor.com/GNUToolchain/
popenv
popd
popenv

# task [078/219] /i686-pc-linux-gnu/toolchain/glibc_headers/glibc_0/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushd /scratch/sandra/nios2-linux-bugfix/obj/glibc-headers-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/default/
make install_root=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./ install-bootstrap-headers=yes install-headers
make csu/subdir_lib
cp csu/crt1.o csu/crti.o csu/crtn.o /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./
popd
/scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-gcc -o /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//libc.so -nostdlib -nostartfiles -shared -x c /dev/null
popenv

# task [079/219] /i686-pc-linux-gnu/toolchain/gcc_second/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenvvar AR_FOR_TARGET nios2-linux-gnu-ar
pushenvvar NM_FOR_TARGET nios2-linux-gnu-nm
pushenvvar OBJDUMP_FOR_TARET nios2-linux-gnu-objdump
pushenvvar STRIP_FOR_TARGET nios2-linux-gnu-strip
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/gcc-second-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/gcc-second-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/gcc-second-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/gcc-nios2 --build=i686-pc-linux-gnu --host=i686-pc-linux-gnu --target=nios2-linux-gnu --enable-threads --disable-libmudflap --disable-libssp --disable-libstdcxx-pch --with-gnu-as --with-gnu-ld '--with-specs=%{save-temps: -fverbose-asm} -D__CS_SOURCERYGXX_MAJ__=2013 -D__CS_SOURCERYGXX_MIN__=5 -D__CS_SOURCERYGXX_REV__=43 %{O2:%{!fno-remove-local-statics: -fremove-local-statics}} %{O*:%{O|O0|O1|O2|Os:;:%{!fno-remove-local-statics: -fremove-local-statics}}}' --enable-languages=c,c++ --enable-shared --enable-lto --enable-symvers=gnu --enable-__cxa_atexit '--with-pkgversion=Sourcery CodeBench Lite 2013.05-43' --with-bugurl=https://sourcery.mentor.com/GNUToolchain/ --disable-nls --prefix=/opt/codesourcery --disable-libffi --disable-libquadmath --disable-libitm --disable-libatomic --enable-languages=c --with-sysroot=/opt/codesourcery/nios2-linux-gnu/libc --with-build-sysroot=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc --with-gmp=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --with-mpfr=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --with-mpc=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --with-ppl=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr '--with-host-libstdcxx=-static-libgcc -Wl,-Bstatic,-lstdc++,-Bdynamic -lm' --with-cloog=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --with-libelf=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --disable-libgomp --disable-libitm --enable-poison-system-directories --with-build-time-tools=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin --with-build-time-tools=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin SED=sed
popd
popenv
popenv

# task [080/219] /i686-pc-linux-gnu/toolchain/gcc_second/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenvvar AR_FOR_TARGET nios2-linux-gnu-ar
pushenvvar NM_FOR_TARGET nios2-linux-gnu-nm
pushenvvar OBJDUMP_FOR_TARET nios2-linux-gnu-objdump
pushenvvar STRIP_FOR_TARGET nios2-linux-gnu-strip
pushd /scratch/sandra/nios2-linux-bugfix/obj/gcc-second-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4 LDFLAGS_FOR_TARGET=--sysroot=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc CPPFLAGS_FOR_TARGET=--sysroot=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc build_tooldir=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu
popd
popenv
popenv

# task [081/219] /i686-pc-linux-gnu/toolchain/gcc_second/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenvvar AR_FOR_TARGET nios2-linux-gnu-ar
pushenvvar NM_FOR_TARGET nios2-linux-gnu-nm
pushenvvar OBJDUMP_FOR_TARET nios2-linux-gnu-objdump
pushenvvar STRIP_FOR_TARGET nios2-linux-gnu-strip
pushd /scratch/sandra/nios2-linux-bugfix/obj/gcc-second-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make prefix=/scratch/sandra/nios2-linux-bugfix/install exec_prefix=/scratch/sandra/nios2-linux-bugfix/install libdir=/scratch/sandra/nios2-linux-bugfix/install/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/man install
make prefix=/scratch/sandra/nios2-linux-bugfix/install exec_prefix=/scratch/sandra/nios2-linux-bugfix/install libdir=/scratch/sandra/nios2-linux-bugfix/install/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/man install-html
make prefix=/scratch/sandra/nios2-linux-bugfix/install exec_prefix=/scratch/sandra/nios2-linux-bugfix/install libdir=/scratch/sandra/nios2-linux-bugfix/install/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/man install-pdf
popd
popenv
popenv

# task [082/219] /i686-pc-linux-gnu/toolchain/gcc_second/postinstall
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenvvar AR_FOR_TARGET nios2-linux-gnu-ar
pushenvvar NM_FOR_TARGET nios2-linux-gnu-nm
pushenvvar OBJDUMP_FOR_TARET nios2-linux-gnu-objdump
pushenvvar STRIP_FOR_TARGET nios2-linux-gnu-strip
pushd /scratch/sandra/nios2-linux-bugfix/install
rm ./lib/libiberty.a
rmdir include
popd
rm -rf /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/gccinstall /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/gcc/gccinstall.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info/gccinstall.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly gccinstall
rm -rf /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/gccint /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/gcc/gccint.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info/gccint.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly gccint
rm -rf /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/cppinternals /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/gcc/cppinternals.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info/cppinternals.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly cppinternals
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/libiberty.html /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/libiberty.pdf
popenv
popenv

# task [083/219] /i686-pc-linux-gnu/toolchain/glibc/glibc_0/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./lib/./ /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/glibc-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/default/
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/glibc-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/default/
pushd /scratch/sandra/nios2-linux-bugfix/obj/glibc-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/default/
pushenv
pushenvvar BUILD_CC gcc
pushenvvar CC '/scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-gcc '
pushenvvar CXX '/scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-g++ '
pushenvvar CFLAGS '-g -O2'
pushenvvar AR /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-ar
pushenvvar NM /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-nm
pushenvvar RANLIB /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-ranlib
pushenvvar READELF /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-readelf
pushenvvar PERL /usr/bin/perl
pushenvvar BASH_SHELL /bin/bash
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/obj/glibc-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/default/ --prefix=/usr --with-headers=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/include --build=i686-pc-linux-gnu --host=nios2-linux-gnu --disable-profile --without-gd --without-cvs --enable-obsolete-rpc --enable-add-ons --enable-kernel=2.6.21 '--with-pkgversion=Sourcery CodeBench Lite 2013.05-43' --with-bugurl=https://sourcery.mentor.com/GNUToolchain/
popenv
popd
popenv

# task [084/219] /i686-pc-linux-gnu/toolchain/glibc/glibc_0/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushd /scratch/sandra/nios2-linux-bugfix/obj/glibc-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/default/
pushenv
pushenvvar PARALLELMFLAGS -j4
make
popenv
popd
popenv

# task [085/219] /i686-pc-linux-gnu/toolchain/glibc/glibc_0/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/info/dir /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/share/info/dir
pushd /scratch/sandra/nios2-linux-bugfix/obj/glibc-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/default/
make install_root=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./ install
make html
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html
copy_dir /scratch/sandra/nios2-linux-bugfix/obj/glibc-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/default//manual/libc /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/libc
make pdf
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf
cp /scratch/sandra/nios2-linux-bugfix/obj/glibc-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/default//manual/libc.pdf /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/libc.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/libexec/pt_chown
rm -rf /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libc_pic
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libBrokenLocale_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libanl_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libc_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libcidn_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libcrypt_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libdl_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libm_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libnsl_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libnss_compat_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libnss_db_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libnss_dns_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libnss_files_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libnss_hesiod_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libnss_nis_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libnss_nisplus_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libresolv_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./librt_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libthread_db_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libutil_pic.a
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libBrokenLocale_pic.map
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libanl_pic.map
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libc_pic.map
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libcidn_pic.map
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libcrypt_pic.map
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libdl_pic.map
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libm_pic.map
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libnsl_pic.map
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libnss_compat_pic.map
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libnss_db_pic.map
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libnss_dns_pic.map
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libnss_files_pic.map
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libnss_hesiod_pic.map
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libnss_nis_pic.map
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libnss_nisplus_pic.map
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libresolv_pic.map
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./librt_pic.map
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libthread_db_pic.map
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./libutil_pic.map
pushd /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv sbin/ldconfig /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv sbin/sln /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/bin/catchsegv /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/bin/gencat /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/bin/getconf /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/bin/getent /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/bin/iconv /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/bin/ldd /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/bin/locale /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/bin/localedef /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/bin/makedb /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/bin/mtrace /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/bin/pcprofiledump /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/bin/pldd /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/bin/rpcgen /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/bin/sotruss /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/bin/sprof /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/bin/tzselect /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/bin/xtrace /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/sbin/iconvconfig /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/sbin/nscd /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/sbin/zdump /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/sbin/zic /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/libexec/getconf/POSIX_V6_ILP32_OFF32 /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/libexec/getconf/POSIX_V6_ILP32_OFFBIG /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/libexec/getconf/POSIX_V7_ILP32_OFF32 /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/libexec/getconf/POSIX_V7_ILP32_OFFBIG /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/libexec/getconf/XBS5_ILP32_OFF32 /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/libexec/getconf/XBS5_ILP32_OFFBIG /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
popd
popd
popenv

# task [086/219] /i686-pc-linux-gnu/toolchain/build_glibc_localedef
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/glibc_localedef-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/glibc_localedef-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/glibc_localedef-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
/scratch/sandra/nios2-linux-bugfix/src/glibc_localedef-trunk-nios2-2_17/configure --prefix=/usr --with-glibc=/scratch/sandra/nios2-linux-bugfix/src/glibc-trunk-nios2-2_17
make -j4
popd
popenv

# task [087/219] /i686-pc-linux-gnu/toolchain/install_glibc_locale/glibc_0
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushd /scratch/sandra/nios2-linux-bugfix/obj/glibc_localedef-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4 install_root=/scratch/sandra/nios2-linux-bugfix/obj/glibc_localedef-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/little4 'LOCALEDEF_OPTS=--little-endian --uint32-align=4' install-locales
mkdir -p little4/usr/lib/locale/test
cat > little4/usr/lib/locale/test/README <<'EOF0'
The locales in this directory are used by the GLIBC test harness
EOF0
pushenv
pushenvvar I18NPATH /scratch/sandra/nios2-linux-bugfix/src/glibc-trunk-nios2-2_17/localedata
pushenvvar GCONV_PATH iconvdata
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/de_DE -f glibc/localedata/charmaps/ISO-8859-1 little4/usr/lib/locale/test/de_DE.ISO-8859-1
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/de_DE -f glibc/localedata/charmaps/UTF-8 little4/usr/lib/locale/test/de_DE.UTF-8
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/en_US -f glibc/localedata/charmaps/ANSI_X3.4-1968 little4/usr/lib/locale/test/en_US.ANSI_X3.4-1968
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/en_US -f glibc/localedata/charmaps/ISO-8859-1 little4/usr/lib/locale/test/en_US.ISO-8859-1
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/en_US -f glibc/localedata/charmaps/UTF-8 little4/usr/lib/locale/test/en_US.UTF-8
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/ja_JP -f glibc/localedata/charmaps/EUC-JP little4/usr/lib/locale/test/ja_JP.EUC-JP
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/da_DK -f glibc/localedata/charmaps/ISO-8859-1 little4/usr/lib/locale/test/da_DK.ISO-8859-1
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/hr_HR -f glibc/localedata/charmaps/ISO-8859-2 little4/usr/lib/locale/test/hr_HR.ISO-8859-2
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/sv_SE -f glibc/localedata/charmaps/ISO-8859-1 little4/usr/lib/locale/test/sv_SE.ISO-8859-1
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/ja_JP -f glibc/localedata/charmaps/SHIFT_JIS little4/usr/lib/locale/test/ja_JP.SJIS
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/fr_FR -f glibc/localedata/charmaps/ISO-8859-1 little4/usr/lib/locale/test/fr_FR.ISO-8859-1
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/nb_NO -f glibc/localedata/charmaps/ISO-8859-1 little4/usr/lib/locale/test/nb_NO.ISO-8859-1
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/nn_NO -f glibc/localedata/charmaps/ISO-8859-1 little4/usr/lib/locale/test/nn_NO.ISO-8859-1
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/tr_TR -f glibc/localedata/charmaps/UTF-8 little4/usr/lib/locale/test/tr_TR.UTF-8
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/cs_CZ -f glibc/localedata/charmaps/UTF-8 little4/usr/lib/locale/test/cs_CZ.UTF-8
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/zh_TW -f glibc/localedata/charmaps/EUC-TW little4/usr/lib/locale/test/zh_TW.EUC-TW
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/fa_IR -f glibc/localedata/charmaps/UTF-8 little4/usr/lib/locale/test/fa_IR.UTF-8
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/fr_FR -f glibc/localedata/charmaps/UTF-8 little4/usr/lib/locale/test/fr_FR.UTF-8
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/ja_JP -f glibc/localedata/charmaps/UTF-8 little4/usr/lib/locale/test/ja_JP.UTF-8
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/si_LK -f glibc/localedata/charmaps/UTF-8 little4/usr/lib/locale/test/si_LK.UTF-8
./localedef --quiet -c --little-endian --uint32-align=4 -i glibc/localedata/locales/tr_TR -f glibc/localedata/charmaps/ISO-8859-9 little4/usr/lib/locale/test/tr_TR.ISO-8859-9
popenv
copy_dir little4 /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./
popd
popenv

# task [088/219] /i686-pc-linux-gnu/toolchain/gcc_final/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenvvar AR_FOR_TARGET nios2-linux-gnu-ar
pushenvvar NM_FOR_TARGET nios2-linux-gnu-nm
pushenvvar OBJDUMP_FOR_TARET nios2-linux-gnu-objdump
pushenvvar STRIP_FOR_TARGET nios2-linux-gnu-strip
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/gcc-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/gcc-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/gcc-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/gcc-nios2 --build=i686-pc-linux-gnu --host=i686-pc-linux-gnu --target=nios2-linux-gnu --enable-threads --disable-libmudflap --disable-libssp --disable-libstdcxx-pch --with-gnu-as --with-gnu-ld '--with-specs=%{save-temps: -fverbose-asm} -D__CS_SOURCERYGXX_MAJ__=2013 -D__CS_SOURCERYGXX_MIN__=5 -D__CS_SOURCERYGXX_REV__=43 %{O2:%{!fno-remove-local-statics: -fremove-local-statics}} %{O*:%{O|O0|O1|O2|Os:;:%{!fno-remove-local-statics: -fremove-local-statics}}}' --enable-languages=c,c++ --enable-shared --enable-lto --enable-symvers=gnu --enable-__cxa_atexit '--with-pkgversion=Sourcery CodeBench Lite 2013.05-43' --with-bugurl=https://sourcery.mentor.com/GNUToolchain/ --disable-nls --prefix=/opt/codesourcery --with-sysroot=/opt/codesourcery/nios2-linux-gnu/libc --with-build-sysroot=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc --with-gmp=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --with-mpfr=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --with-mpc=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --with-ppl=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr '--with-host-libstdcxx=-static-libgcc -Wl,-Bstatic,-lstdc++,-Bdynamic -lm' --with-cloog=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --with-libelf=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --disable-libgomp --disable-libitm --enable-poison-system-directories --with-build-time-tools=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin --with-build-time-tools=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin SED=sed
popd
popenv
popenv

# task [089/219] /i686-pc-linux-gnu/toolchain/gcc_final/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenvvar AR_FOR_TARGET nios2-linux-gnu-ar
pushenvvar NM_FOR_TARGET nios2-linux-gnu-nm
pushenvvar OBJDUMP_FOR_TARET nios2-linux-gnu-objdump
pushenvvar STRIP_FOR_TARGET nios2-linux-gnu-strip
pushd /scratch/sandra/nios2-linux-bugfix/obj/gcc-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4 LDFLAGS_FOR_TARGET=--sysroot=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc CPPFLAGS_FOR_TARGET=--sysroot=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc build_tooldir=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu
popd
popenv
popenv

# task [090/219] /i686-pc-linux-gnu/toolchain/gcc_final/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenvvar AR_FOR_TARGET nios2-linux-gnu-ar
pushenvvar NM_FOR_TARGET nios2-linux-gnu-nm
pushenvvar OBJDUMP_FOR_TARET nios2-linux-gnu-objdump
pushenvvar STRIP_FOR_TARGET nios2-linux-gnu-strip
pushd /scratch/sandra/nios2-linux-bugfix/obj/gcc-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make prefix=/scratch/sandra/nios2-linux-bugfix/install exec_prefix=/scratch/sandra/nios2-linux-bugfix/install libdir=/scratch/sandra/nios2-linux-bugfix/install/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/man install
make prefix=/scratch/sandra/nios2-linux-bugfix/install exec_prefix=/scratch/sandra/nios2-linux-bugfix/install libdir=/scratch/sandra/nios2-linux-bugfix/install/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/man install-html
make prefix=/scratch/sandra/nios2-linux-bugfix/install exec_prefix=/scratch/sandra/nios2-linux-bugfix/install libdir=/scratch/sandra/nios2-linux-bugfix/install/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/man install-pdf
popd
popenv
popenv

# task [091/219] /i686-pc-linux-gnu/toolchain/gcc_final/postinstall
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenvvar AR_FOR_TARGET nios2-linux-gnu-ar
pushenvvar NM_FOR_TARGET nios2-linux-gnu-nm
pushenvvar OBJDUMP_FOR_TARET nios2-linux-gnu-objdump
pushenvvar STRIP_FOR_TARGET nios2-linux-gnu-strip
pushd /scratch/sandra/nios2-linux-bugfix/install
rm ./lib/libiberty.a
rmdir include
popd
rm -rf /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/gccinstall /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/gcc/gccinstall.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info/gccinstall.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly gccinstall
rm -rf /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/gccint /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/gcc/gccint.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info/gccint.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly gccint
rm -rf /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/cppinternals /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/gcc/cppinternals.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info/cppinternals.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly cppinternals
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/libiberty.html /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/libiberty.pdf
rm -rf /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/lib/.//libstdc++.so.6.0.17-gdb.py
popenv
popenv

# task [092/219] /i686-pc-linux-gnu/toolchain/zlib/0/copy
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/zlib-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
copy_dir_clean /scratch/sandra/nios2-linux-bugfix/src/zlib-1.2.7 /scratch/sandra/nios2-linux-bugfix/obj/zlib-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
chmod -R u+w /scratch/sandra/nios2-linux-bugfix/obj/zlib-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
popenv

# task [093/219] /i686-pc-linux-gnu/toolchain/zlib/0/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushd /scratch/sandra/nios2-linux-bugfix/obj/zlib-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushenv
pushenvvar CFLAGS '-O3 -fPIC'
pushenvvar CC 'i686-pc-linux-gnu-gcc  '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
./configure --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --static
popenv
popd
popenv

# task [094/219] /i686-pc-linux-gnu/toolchain/zlib/0/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushd /scratch/sandra/nios2-linux-bugfix/obj/zlib-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4
popd
popenv

# task [095/219] /i686-pc-linux-gnu/toolchain/zlib/0/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushd /scratch/sandra/nios2-linux-bugfix/obj/zlib-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make install
popd
popenv

# task [096/219] /i686-pc-linux-gnu/toolchain/expat/0/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/expat-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/expat-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/expat-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/expat-2013.05 --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --disable-shared --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-pc-linux-gnu --disable-nls
popd
popenv
popenv
popenv

# task [097/219] /i686-pc-linux-gnu/toolchain/expat/0/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/expat-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4
popd
popenv
popenv
popenv

# task [098/219] /i686-pc-linux-gnu/toolchain/expat/0/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/expat-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make install
popd
popenv
popenv
popenv

# task [099/219] /i686-pc-linux-gnu/toolchain/ncurses/0/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/ncurses-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/ncurses-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/ncurses-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/ncurses-5.9 --prefix=/usr --disable-shared --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-pc-linux-gnu --without-debug --without-cxx --without-ada --disable-nls
popd
popenv
popenv
popenv

# task [100/219] /i686-pc-linux-gnu/toolchain/ncurses/0/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/ncurses-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j1
popd
popenv
popenv
popenv

# task [101/219] /i686-pc-linux-gnu/toolchain/ncurses/0/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/ncurses-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make install DESTDIR=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu
popd
popenv
popenv
popenv

# task [102/219] /i686-pc-linux-gnu/toolchain/gdb/0/copy
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/lib
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/gdb-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
copy_dir_clean /scratch/sandra/nios2-linux-bugfix/src/gdb-nios2 /scratch/sandra/nios2-linux-bugfix/obj/gdb-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
chmod -R u+w /scratch/sandra/nios2-linux-bugfix/obj/gdb-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
touch /scratch/sandra/nios2-linux-bugfix/obj/gdb-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/.gnu-stamp
popenv
popenv
popenv

# task [103/219] /i686-pc-linux-gnu/toolchain/gdb/0/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/lib
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/gdb-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/gdb-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/gdb-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/obj/gdb-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu --prefix=/opt/codesourcery --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-pc-linux-gnu '--with-pkgversion=Sourcery CodeBench Lite 2013.05-43' --with-bugurl=https://sourcery.mentor.com/GNUToolchain/ --disable-libmcheck --disable-nls --with-libexpat-prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --with-system-gdbinit=/opt/codesourcery/i686-pc-linux-gnu/nios2-linux-gnu/lib/gdbinit '--with-gdb-datadir='\''${prefix}'\''/nios2-linux-gnu/share/gdb'
popd
popenv
popenv
popenv

# task [104/219] /i686-pc-linux-gnu/toolchain/gdb/0/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/lib
pushd /scratch/sandra/nios2-linux-bugfix/obj/gdb-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4
popd
popenv
popenv
popenv

# task [105/219] /i686-pc-linux-gnu/toolchain/gdb/0/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/lib
pushd /scratch/sandra/nios2-linux-bugfix/obj/gdb-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make install prefix=/scratch/sandra/nios2-linux-bugfix/install exec_prefix=/scratch/sandra/nios2-linux-bugfix/install libdir=/scratch/sandra/nios2-linux-bugfix/install/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/man datadir=/scratch/sandra/nios2-linux-bugfix/install/share
popd
popenv
popenv
popenv

# task [106/219] /i686-pc-linux-gnu/toolchain/gdb/0/postinstall
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr/lib
pushd /scratch/sandra/nios2-linux-bugfix/obj/gdb-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make prefix=/scratch/sandra/nios2-linux-bugfix/install exec_prefix=/scratch/sandra/nios2-linux-bugfix/install libdir=/scratch/sandra/nios2-linux-bugfix/install/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/man datadir=/scratch/sandra/nios2-linux-bugfix/install/share install-html
make prefix=/scratch/sandra/nios2-linux-bugfix/install exec_prefix=/scratch/sandra/nios2-linux-bugfix/install libdir=/scratch/sandra/nios2-linux-bugfix/install/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/man datadir=/scratch/sandra/nios2-linux-bugfix/install/share install-pdf
popd
rm -rf /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/annotate /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/annotate.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info/annotate.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly annotate
rm -rf /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/gdbint /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/gdbint.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info/gdbint.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly gdbint
rm -rf /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/stabs /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/stabs.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info/stabs.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly stabs
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/configure.html /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/etc/configure.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info/configure.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly configure
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/standards.html /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/etc/standards.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info/standards.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly standards
rmdir /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/etc
rm -rf /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/bfd.html /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/bfd.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info/bfd.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly bfd
rm -f /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html/libiberty.html /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf/libiberty.pdf
pushd /scratch/sandra/nios2-linux-bugfix/install
mkdir -p i686-pc-linux-gnu/nios2-linux-gnu/include/
mv include/gdb i686-pc-linux-gnu/nios2-linux-gnu/include
rmdir include
popd
popenv
popenv
popenv

# task [107/219] /i686-pc-linux-gnu/toolchain/build_gdbserver/glibc_0
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/gdbserver-2013.05-43-nios2-linux-gnu/default/
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/gdbserver-2013.05-43-nios2-linux-gnu/default/
pushd /scratch/sandra/nios2-linux-bugfix/obj/gdbserver-2013.05-43-nios2-linux-gnu/default/
pushenv
pushenvvar RANLIB nios2-linux-gnu-ranlib
pushenvvar CC 'nios2-linux-gnu-gcc '
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/gdb-nios2/gdb/gdbserver --build=i686-pc-linux-gnu --host=nios2-linux-gnu --prefix=/usr --disable-shared '--with-pkgversion=Sourcery CodeBench Lite 2013.05-43' --with-bugurl=https://sourcery.mentor.com/GNUToolchain/ --disable-libmcheck
popenv
make -j4
make DESTDIR=/scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./ install
pushd /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
mv usr/bin/gdbserver /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./bin
popd
popd
popenv

# task [108/219] /i686-pc-linux-gnu/toolchain/cs3/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenvvar CC nios2-linux-gnu-gcc
pushenvvar CXX nios2-linux-gnu-g++
pushenvvar AR nios2-linux-gnu-ar
pushenvvar RANLIB nios2-linux-gnu-ranlib
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/cs3-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/cs3-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/cs3-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/cs3-2013.05 --prefix=/opt/codesourcery --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=nios2-linux-gnu --disable-cs3 '--with-pkgversion=Sourcery CodeBench Lite 2013.05-43' --with-bugurl=https://sourcery.mentor.com/GNUToolchain/ --disable-nls --with-intermediate-dir=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/cs3 '--with-configs=public-*'
popd
popenv
popenv
popenv

# task [109/219] /i686-pc-linux-gnu/toolchain/cs3/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenvvar CC nios2-linux-gnu-gcc
pushenvvar CXX nios2-linux-gnu-g++
pushenvvar AR nios2-linux-gnu-ar
pushenvvar RANLIB nios2-linux-gnu-ranlib
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/cs3-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4
popd
popenv
popenv
popenv

# task [110/219] /i686-pc-linux-gnu/toolchain/cs3/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenvvar CC nios2-linux-gnu-gcc
pushenvvar CXX nios2-linux-gnu-g++
pushenvvar AR nios2-linux-gnu-ar
pushenvvar RANLIB nios2-linux-gnu-ranlib
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/cs3-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make install prefix=/scratch/sandra/nios2-linux-bugfix/install exec_prefix=/scratch/sandra/nios2-linux-bugfix/install libdir=/scratch/sandra/nios2-linux-bugfix/install/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/man datadir=/scratch/sandra/nios2-linux-bugfix/install/share
popd
popenv
popenv
popenv

# task [111/219] /i686-pc-linux-gnu/getting_started/copy
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/getting_started-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
copy_dir_clean /scratch/sandra/nios2-linux-bugfix/src/getting-started-2013.05 /scratch/sandra/nios2-linux-bugfix/obj/getting_started-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
chmod -R u+w /scratch/sandra/nios2-linux-bugfix/obj/getting_started-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
touch /scratch/sandra/nios2-linux-bugfix/obj/getting_started-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/.gnu-stamp
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/getting_started-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/sgxx/2013.05
copy_dir_clean /scratch/sandra/nios2-linux-bugfix/src/release-notes-trunk /scratch/sandra/nios2-linux-bugfix/obj/getting_started-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/release-notes
cat > /scratch/sandra/nios2-linux-bugfix/obj/getting_started-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/sgxx/2013.05/nios2-linux-gnu-lite-libs.xml <<'EOF0'
<!DOCTYPE getting-started SYSTEM "getting-started.dtd">
<!-- automatically generated, do not edit -->
<section id="sec-multilibs">
  <title>Library Configurations</title>
  <para>
    &csl_prod; for &csl_target_name; includes the following library
    configuration.
  </para>
  <informaltable>
    <?dbfo keep-together="always" ?>
    <?dbfo table-width="100%" ?>
    <?dbhtml table-width="100%" ?>
    <tgroup cols="2" align="left">
      <colspec colname="c1" colwidth="1*" />
      <colspec colname="c2" colwidth="2*" />
      <thead>
	<row>
	  <entry namest="c1" nameend="c2">Nios II - GLIBC</entry>
	</row>
      </thead>
      <tbody>
	<row>
	  <entry>Command-line option(s):</entry>
	  <entry>default</entry>
	</row>
	<row>
	  <entry>Sysroot subdirectory:</entry>
	  <entry><filename>./</filename></entry>
	</row>
	<row>
	  <entry>Dynamic linker:</entry>
	  <entry><filename>lib/ld.so.1</filename></entry>
	</row>
      </tbody>
    </tgroup>
  </informaltable>
  <xi:include href="/scratch/sandra/nios2-linux-bugfix/obj/getting_started-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/library-intro.xml" xpointer="xpointer(*/*)"/>
</section>
EOF0
cat > /scratch/sandra/nios2-linux-bugfix/obj/getting_started-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/sgxx/2013.05/nios2-linux-gnu-lite.dtd <<'EOF0'
<!-- automatically generated, do not edit -->
<!ENTITY csl_multilib_sec "<xi:include href='/scratch/sandra/nios2-linux-bugfix/obj/getting_started-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/sgxx/2013.05/nios2-linux-gnu-lite-libs.xml'/>">
<!ENTITY csl_default_ldso_dirname "lib">
<!ENTITY csl_default_ldso_basename "ld.so.1">
<!ENTITY csl_min_kernel_version "2.6.21">
<!ENTITY csl_subpackage_sec "">
<!ENTITY csl_binutils_component_license "&csl_gpl3.0_license;">
<!ENTITY csl_binutils_component_version "2.23.52">
<!ENTITY csl_gcc_component_license "&csl_gpl3.0_license;">
<!ENTITY csl_gcc_component_version "4.7.3">
<!ENTITY csl_gdb_component_license "&csl_gpl3.0_license;">
<!ENTITY csl_gdb_component_version "7.4.50">
<!ENTITY csl_cs3_component_license "&csl_cs_license;">
<!ENTITY csl_sprite_component_license "&csl_cs_license;">
<!ENTITY csl_glibc_component_license "&csl_lgpl2.1_license;">
<!ENTITY csl_glibc_component_version "trunk-nios2-2_17">
<!ENTITY csl_linux_component_license "&csl_gpl2.0_license;">
<!ENTITY csl_linux_component_version "nios2">
<!ENTITY csl_make_component_license "&csl_gpl2.0_license;">
<!ENTITY csl_coreutils_component_license "&csl_gpl2.0_license;">
EOF0
popenv
popenv
popenv

# task [112/219] /i686-pc-linux-gnu/getting_started/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/getting_started-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/getting_started-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/getting_started-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/obj/getting_started-src-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu --prefix=/opt/codesourcery --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-pc-linux-gnu --disable-nls --with-release-config=sgxx/2013.05/nios2-linux-gnu-lite '--with-components=binutils gcc gdb zlib gmp mpfr mpc cloog ppl ncurses getting_started installanywhere cs3 sprite glibc glibc_localedef linux coremark branding contents comparison csl_tests dejagnu_boards scripts xfails portal harness config libiconv libelf expat csl_docbook release_notes make coreutils gdil edge_utils eglibc_configs no-alf no-alp_examples no-android no-antlr no-apm_examples no-argp no-arm_stub no-ashling no-atlas_install no-board_support no-brewmp no-ccs no-ccs_mdi_library no-cctools no-cfpe_stub no-cgen no-cml no-codewarrior_plugin no-com_codesourcery_util no-cs no-csibe no-csldoc no-cslibc_manual no-cula_install no-cxxabi no-cygwin_wrapper no-dbgrtool no-dhrystone no-dinkum no-dltk no-eclipse no-eclipse_cdt no-eclipse_common no-eclipse_patches no-eclipse_plugin no-eclipse_zip no-eembc no-eglibc no-elf2flt no-emf no-fbench no-fftw no-fftw_source no-flexlm_utils no-gettext no-glib no-glibc_ports no-harness_test no-hexagon_codec no-hexagon_dinkum no-ide no-intel_sprite no-isl no-json_c no-kinetis_examples no-libcsftdi no-libcurl no-libffi no-libftdi no-libusb no-libusb_win32 no-license no-license_lib no-lsb no-mep_cs no-mepl no-mesa no-mesp_docs no-mettools no-mgls no-mingw no-mips_sprite no-mips_toolchain_manual no-mklibs no-msvcrt no-newlib no-nucleus_binary_demos no-ocdremote no-omf2elf no-openposix no-openssl no-oprofile no-oprofile_plugin no-osawareness_plugin no-p2_installer no-p2_repogenerator no-ph no-plex_skel no-popt no-power_sprite no-prelink no-python no-python_hexagon_linux no-python_macosx no-python_win32 no-qemu no-qmtest_ph no-redsea no-rpm no-sgxx_application_notes no-sgxx_demos no-simdmath no-spec2000 no-spec2000_configs no-spec2006 no-spec2xxx no-stellarisware no-stm32_generate_sgxx no-stm32_stdperiph_lib no-stm32_usb_fs_device_lib no-svxx_api_reference no-svxx_benchmarks no-svxx_binary no-svxx_copy_variants no-svxx_documentation no-svxx_fftw no-svxx_source no-sysroot_utils no-sysrootinfo no-trace_processor no-uclibc no-uclibc_configs no-verifone_plugin no-w32api no-windows_script_wrapper no-xlp_mdi no-xulrunner' '--with-features=glibc_armhf_legacy_dyn_linker sgxx_version gdbserver tarball public elf elf_shlibs elf_sysroot sysroot linux_kernel ldso_library_path_option sprite usbblaster' '--with-hosts=i686-pc-linux-gnu i686-mingw32' '--with-target-arch-name=Nios II' --with-target-os-name=GNU/Linux --with-intermediate-dir=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/getting_started --with-csl-docbook=/scratch/sandra/nios2-linux-bugfix/src/csl-docbook-trunk --with-version-string=2013.05-43 '--with-pkgversion=Sourcery CodeBench Lite 2013.05-43' '--with-brand=Sourcery CodeBench Lite' --with-pkg_prefix=sourceryg++ '--with-xml-catalog-files=/usr/local/tools/gcc-4.7.3/share/sgml/docbook/docbook-xsl/catalog.xml /etc/xml/catalog' --with-license=lite --with-cs3=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/cs3/doc
popd
popenv
popenv
popenv

# task [113/219] /i686-pc-linux-gnu/getting_started/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/getting_started-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4
popd
popenv
popenv
popenv

# task [114/219] /i686-pc-linux-gnu/getting_started/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/getting_started-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make install prefix=/scratch/sandra/nios2-linux-bugfix/install exec_prefix=/scratch/sandra/nios2-linux-bugfix/install libdir=/scratch/sandra/nios2-linux-bugfix/install/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/man datadir=/scratch/sandra/nios2-linux-bugfix/install/share docdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu
popd
popenv
popenv
popenv

# task [115/219] /i686-pc-linux-gnu/csl_docbook
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
cp /scratch/sandra/nios2-linux-bugfix/src/csl-docbook-trunk/css/cs.css /scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html
popenv

# task [116/219] /i686-pc-linux-gnu/gdil/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/gdil-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/gdil-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/gdil-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/gdil-2013.05 --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --disable-shared --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-pc-linux-gnu --with-csl-docbook=/scratch/sandra/nios2-linux-bugfix/src/csl-docbook-trunk '--with-xml-catalog-files=/usr/local/tools/gcc-4.7.3/share/sgml/docbook/docbook-xsl/catalog.xml /etc/xml/catalog' --disable-nls --with-expat=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr '--with-features=xml autoconf driver config sprite doc'
popd
popenv
popenv
popenv

# task [117/219] /i686-pc-linux-gnu/gdil/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/gdil-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4
popd
popenv
popenv
popenv

# task [118/219] /i686-pc-linux-gnu/gdil/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/gdil-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make install
popd
popenv
popenv
popenv

# task [119/219] /i686-pc-linux-gnu/edge_utils/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/edge-utils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/edge-utils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
copy_dir_clean /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source /scratch/sandra/nios2-linux-bugfix/obj/edge-utils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/source
chmod -R u+w /scratch/sandra/nios2-linux-bugfix/obj/edge-utils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -C /scratch/sandra/nios2-linux-bugfix/obj/edge-utils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/source/debugger/utils -j4 BLDTYPE=rls HOST_OS=lnx GPREF=i686-pc-linux-gnu- lnx_rls/libUtils.a
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/edge-utils-i686-pc-linux-gnu/lib
cp /scratch/sandra/nios2-linux-bugfix/obj/edge-utils-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu/source/debugger/utils/lnx_rls/libUtils.a /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/edge-utils-i686-pc-linux-gnu/lib
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/edge-utils-i686-pc-linux-gnu/include
cp /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/NumericValue.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/clBaseTypes.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/clByteBuffer.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/clErrorLogger.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/clException.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/clStl.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluAllocators.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluAtomic.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluCLibrary.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluCond.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluConfigData.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluConfigFile.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluDebugTrace.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluDll.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluEmo.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluEndian.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluErrors.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluEvent.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluEventWaiter.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluFile.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluFileUtils.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluFindFile.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluFunctionTrace.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluGuard.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluHighResTimer.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluHtmlLogFile.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluInetPackets.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluInetSocket.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluLogFile.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluMutex.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluOutput.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluQ.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluRedirect.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluRegex.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluSemaphore.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluSerialPort.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluSerialPortBase.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluSerialPortPosix.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluSerialPortWin32.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluSmartPtr.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluSocketBase.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluSplitPath.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluStats.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluStringHeapManager.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluStringUtil.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluSync.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluTask.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluTextTable.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluThread.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluThreadPool.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluThreadedMember.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluTime.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluTimerQueue.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluTransaction.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluTransactionIdMap.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluUniqueIDManager.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluUnixSignalHandling.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluUtils.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluUtilsPch.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluValueMap.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluVector.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluXmlParser.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/rpn_calc.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/stdint_opus.h /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/edge-utils-i686-pc-linux-gnu/include
cp /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluConfigData.inl /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluEvent.inl /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluGuard.inl /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluHighResTimer.inl /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluLogFile.inl /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluMutex.inl /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluTask.inl /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluThread.inl /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluTime.inl /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/edge-utils-i686-pc-linux-gnu/include
popenv

# task [120/219] /i686-pc-linux-gnu/sprite/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/sprite-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/sprite-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/sprite-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/sprite-2013.05 --prefix=/opt/codesourcery --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-pc-linux-gnu '--with-pkgversion=Sourcery CodeBench Lite 2013.05-43' --with-bugurl=https://sourcery.mentor.com/GNUToolchain/ --enable-backends=usbblaster --disable-nls --with-gdil=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --with-expat=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-pc-linux-gnu/usr --with-boards=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/cs3/boards --with-edge-utils=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/edge-utils-i686-pc-linux-gnu --with-mep_cs=
popd
popenv
popenv
popenv

# task [121/219] /i686-pc-linux-gnu/sprite/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/sprite-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make -j4
popd
popenv
popenv
popenv

# task [122/219] /i686-pc-linux-gnu/sprite/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/sprite-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu
make install prefix=/scratch/sandra/nios2-linux-bugfix/install exec_prefix=/scratch/sandra/nios2-linux-bugfix/install libdir=/scratch/sandra/nios2-linux-bugfix/install/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/share/doc/sourceryg++-nios2-linux-gnu/man datadir=/scratch/sandra/nios2-linux-bugfix/install/share
popd
popenv
popenv
popenv

# task [123/219] /i686-pc-linux-gnu/sprite/postinstall
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushenv
pushenv
popenv
popenv
popenv

# task [124/219] /i686-pc-linux-gnu/finalize_libc_installation
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
cp -d /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/lib/libgcc_s.so /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./lib/./
cp -d /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/lib/libgcc_s.so.1 /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./lib/./
cp -d /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/lib/./libgcc_s.so /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./lib/./
cp -d /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/lib/./libgcc_s.so.1 /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./lib/./
cp -d /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/lib/./libstdc++.so /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./
cp -d /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/lib/./libstdc++.so.6 /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./
cp -d /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/lib/./libstdc++.so.6.0.17 /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./
cp -d /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/lib/./libstdc++.a /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/./
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/gdbserver /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin/gdbserver
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./sbin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/ldconfig /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./sbin/ldconfig
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./sbin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/sln /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./sbin/sln
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/catchsegv /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin/catchsegv
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/gencat /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin/gencat
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/getconf /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin/getconf
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/getent /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin/getent
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/iconv /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin/iconv
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/ldd /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin/ldd
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/locale /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin/locale
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/localedef /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin/localedef
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/makedb /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin/makedb
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/mtrace /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin/mtrace
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/pcprofiledump /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin/pcprofiledump
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/pldd /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin/pldd
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/rpcgen /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin/rpcgen
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/sotruss /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin/sotruss
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/sprof /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin/sprof
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/tzselect /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin/tzselect
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/xtrace /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/bin/xtrace
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/sbin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/iconvconfig /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/sbin/iconvconfig
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/sbin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/nscd /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/sbin/nscd
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/sbin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/zdump /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/sbin/zdump
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/sbin
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/zic /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/sbin/zic
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/libexec/getconf
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/POSIX_V7_ILP32_OFF32 /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/libexec/getconf/POSIX_V7_ILP32_OFF32
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/libexec/getconf
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/POSIX_V7_ILP32_OFFBIG /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/libexec/getconf/POSIX_V7_ILP32_OFFBIG
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/libexec/getconf
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/POSIX_V6_ILP32_OFF32 /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/libexec/getconf/POSIX_V6_ILP32_OFF32
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/libexec/getconf
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/POSIX_V6_ILP32_OFFBIG /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/libexec/getconf/POSIX_V6_ILP32_OFFBIG
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/libexec/getconf
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/XBS5_ILP32_OFF32 /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/libexec/getconf/XBS5_ILP32_OFF32
mkdir -p /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/libexec/getconf
ln -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/lib/.//bin/XBS5_ILP32_OFFBIG /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/./usr/libexec/getconf/XBS5_ILP32_OFFBIG
popenv

# task [125/219] /i686-pc-linux-gnu/pretidy_installation
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushd /scratch/sandra/nios2-linux-bugfix/install
rm ./lib/libiberty.a
popd
popenv

# task [126/219] /i686-pc-linux-gnu/remove_libtool_archives
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
find /scratch/sandra/nios2-linux-bugfix/install -name '*.la' -exec rm '{}' ';'
popenv

# task [127/219] /i686-pc-linux-gnu/remove_copied_libs
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
rm -f /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/lib/./libgcc_s.so /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/lib/./libgcc_s.so.1 /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/lib/./libstdc++.so /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/lib/./libstdc++.so.6 /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/lib/./libstdc++.so.6.0.17 /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/lib/./libstdc++.a
popenv

# task [128/219] /i686-pc-linux-gnu/remove_fixed_headers
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
pushd /scratch/sandra/nios2-linux-bugfix/install/lib/gcc/nios2-linux-gnu/4.7.3/include-fixed
popd
popenv

# task [129/219] /i686-pc-linux-gnu/add_tooldir_readme
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
cat > /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin/README.txt <<'EOF0'
The executables in this directory are for internal use by the compiler
and may not operate correctly when used directly.  This directory
should not be placed on your PATH.  Instead, you should use the
executables in ../../bin/ and place that directory on your PATH.
EOF0
popenv

# task [130/219] /i686-pc-linux-gnu/strip_host_objects
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-addr2line
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-ar
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-as
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-c++
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-c++filt
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-cpp
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-elfedit
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-g++
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-gcc
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-gcc-4.7.3
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-gcc-ar
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-gcc-nm
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-gcc-ranlib
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-gcov
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-gdb
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-gprof
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-ld
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-nm
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-objcopy
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-objdump
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-ranlib
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-readelf
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-size
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-sprite
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-strings
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/bin/nios2-linux-gnu-strip
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin/ar
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin/as
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin/c++
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin/g++
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin/gcc
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin/ld
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin/nm
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin/objcopy
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin/objdump
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin/ranlib
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/bin/strip
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/libexec/gcc/nios2-linux-gnu/4.7.3/cc1
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/libexec/gcc/nios2-linux-gnu/4.7.3/collect2
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/libexec/gcc/nios2-linux-gnu/4.7.3/install-tools/fixincl
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/libexec/gcc/nios2-linux-gnu/4.7.3/cc1plus
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/libexec/gcc/nios2-linux-gnu/4.7.3/lto-wrapper
/usr/local/tools/gcc-4.7.3/bin/i686-pc-linux-gnu-strip /scratch/sandra/nios2-linux-bugfix/install/libexec/gcc/nios2-linux-gnu/4.7.3/lto1
popenv

# task [131/219] /i686-pc-linux-gnu/strip_target_objects
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libieee.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libnss_compat.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP1251.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO-2022-JP.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1144.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM9030.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM4899.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1130.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP1254.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO8859-16.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/UTF-32.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/SJIS.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO8859-15.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM918.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO-2022-JP-3.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP1255.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1132.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM870.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM921.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISIRI-3342.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM281.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM500.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/BIG5HKSCS.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/GEORGIAN-PS.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/PT154.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO8859-9E.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM860.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM869.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/UTF-7.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO8859-9.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM891.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/libJISX0213.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EUC-TW.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM861.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM905.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO646.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM277.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO-IR-197.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM902.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP770.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM922.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1122.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM9066.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EBCDIC-AT-DE-A.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1046.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EBCDIC-CA-FR.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/TCVN5712-1.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CSN_369103.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/LATIN-GREEK-1.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM5347.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM937.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/KOI8-U.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM424.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1142.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO-IR-209.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1133.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ARMSCII-8.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/TIS-620.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP737.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1163.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM275.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/HP-ROMAN8.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1145.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO_11548-1.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1167.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM933.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM901.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/NATS-DANO.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1124.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM868.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/HP-THAI8.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM037.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO_6937.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EUC-CN.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/GEORGIAN-ACADEMY.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EBCDIC-US.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1164.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EBCDIC-UK.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM280.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM943.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1149.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/GBK.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO_2033.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EBCDIC-FI-SE.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1143.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO8859-13.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/DEC-MCS.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM4517.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1390.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM4971.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/MIK.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1388.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO8859-7.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/libJIS.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/KOI8-R.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/INIS-CYRILLIC.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EBCDIC-PT.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/INIS-8.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM932.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM904.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1161.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/JOHAB.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO_5428.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP1252.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM935.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO8859-5.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1154.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP1256.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/GB18030.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1148.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/BIG5.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1160.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/LATIN-GREEK.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EBCDIC-DK-NO.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1371.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM290.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM864.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO8859-1.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EBCDIC-ES.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1097.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO8859-11.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EUC-JP.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO8859-2.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM863.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO-2022-CN-EXT.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM939.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM12712.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1155.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/MAC-CENTRALEUROPE.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM038.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM803.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM256.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM875.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/INIS.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1112.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM9448.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1156.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM851.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EUC-KR.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1008_420.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM16804.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1399.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1129.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO_6937-2.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/MAC-UK.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/libGB.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/GREEK7-OLD.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/HP-GREEK8.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO8859-10.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP932.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM856.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1158.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP774.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EBCDIC-ES-A.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM437.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/KOI-8.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO8859-6.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM274.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1157.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM852.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/GREEK-CCITT.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM930.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP1125.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO_5427.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/T.61.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1166.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP1253.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM866.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/NATS-SEFI.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ECMA-CYRILLIC.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM850.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/MACINTOSH.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM903.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1364.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP1258.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/BRF.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1026.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM871.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/SAMI-WS2.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1137.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/libCNS.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EBCDIC-IS-FRISS.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1004.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP1257.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/HP-TURKISH8.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/VISCII.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP771.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1147.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/HP-ROMAN9.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1162.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM866NAV.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CWI.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/TSCII.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM857.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP772.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP10007.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ASMO_449.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO8859-8.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO8859-14.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/MAC-IS.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EBCDIC-ES-S.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO_10367-BOX.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EBCDIC-FR.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO_5427-EXT.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM297.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EBCDIC-AT-DE.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EBCDIC-FI-SE-A.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/KOI8-T.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/MAC-SAMI.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1153.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM285.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/GREEK7.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/RK1048.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM4909.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM874.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EUC-JISX0213.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM862.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EBCDIC-DK-NO-A.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/GBBIG5.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1008.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/SHIFT_JISX0213.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1123.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/KOI8-RU.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/UTF-16.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM278.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EUC-JP-MS.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM284.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM855.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO8859-4.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP775.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1140.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/libISOIR165.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1025.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1047.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM865.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM420.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/GBGBK.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/UNICODE.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM423.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO8859-3.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ANSI_X3.110.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/EBCDIC-IT.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/GOST_19768-74.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP773.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1146.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM1141.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM273.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO-2022-KR.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IEC_P27-1.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/UHC.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/CP1250.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/ISO-2022-CN.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/libKSC.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gconv/IBM880.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libcrypt.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/librt.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/librt.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/gcrt1.o || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libmcheck.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libm.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libc.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libbsd-compat.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libnss_db.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libc.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libstdc++.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libstdc++.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libutil.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libBrokenLocale.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libresolv.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libanl.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libnsl.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libutil.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libdl.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libc_nonshared.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libthread_db.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libBrokenLocale.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/crtn.o || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/audit/sotruss-lib.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/Mcrt1.o || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libnss_files.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/librpcsvc.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libcidn.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libpthread.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/crt1.o || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libg.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libnss_nis.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libpthread_nonshared.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/Scrt1.o || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libresolv.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libnsl.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libanl.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libnss_nisplus.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libnss_hesiod.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libdl.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/crti.o || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libnss_dns.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libpthread.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libcrypt.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/usr/lib/libm.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libpcprofile.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libnss_dns-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libSegFault.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libthread_db-1.0.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libnss_files-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libnss_hesiod-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libdl-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libmemusage.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/librt-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libnss_nisplus-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libnss_db-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libnss_compat-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libpthread-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/ld-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libnsl-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libnss_nis-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libutil-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libm-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libcidn-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libBrokenLocale-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libanl-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libcrypt-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libresolv-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libc-2.17.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/lib/libgcc_s.so || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/lib/libsupc++.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/lib/gcc/nios2-linux-gnu/4.7.3/crtbeginT.o || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/lib/gcc/nios2-linux-gnu/4.7.3/crtbegin.o || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/lib/gcc/nios2-linux-gnu/4.7.3/libgcc.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/lib/gcc/nios2-linux-gnu/4.7.3/libgcc_eh.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/lib/gcc/nios2-linux-gnu/4.7.3/libgcov.a || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/lib/gcc/nios2-linux-gnu/4.7.3/crtn.o || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/lib/gcc/nios2-linux-gnu/4.7.3/crtendS.o || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/lib/gcc/nios2-linux-gnu/4.7.3/crtend.o || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/lib/gcc/nios2-linux-gnu/4.7.3/crti.o || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/lib/gcc/nios2-linux-gnu/4.7.3/crtbeginS.o || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//sbin/ldconfig || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//sbin/sln || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/bin/catchsegv || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/bin/gdbserver || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/bin/gencat || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/bin/getconf || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/bin/getent || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/bin/iconv || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/bin/ldd || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/bin/locale || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/bin/localedef || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/bin/makedb || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/bin/mtrace || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/bin/pcprofiledump || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/bin/pldd || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/bin/rpcgen || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/bin/sotruss || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/bin/sprof || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/bin/tzselect || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/bin/xtrace || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/sbin/iconvconfig || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/sbin/nscd || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/sbin/zdump || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/sbin/zic || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/libexec/getconf/POSIX_V6_ILP32_OFF32 || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/libexec/getconf/POSIX_V6_ILP32_OFFBIG || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/libexec/getconf/POSIX_V7_ILP32_OFF32 || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/libexec/getconf/POSIX_V7_ILP32_OFFBIG || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/libexec/getconf/XBS5_ILP32_OFF32 || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/libexec/getconf/XBS5_ILP32_OFFBIG || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/POSIX_V6_ILP32_OFF32 || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/POSIX_V6_ILP32_OFFBIG || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/POSIX_V7_ILP32_OFF32 || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/POSIX_V7_ILP32_OFFBIG || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/XBS5_ILP32_OFF32 || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/XBS5_ILP32_OFFBIG || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/catchsegv || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/gdbserver || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/gencat || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/getconf || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/getent || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/iconv || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/iconvconfig || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/ldconfig || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/ldd || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/locale || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/localedef || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/makedb || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/mtrace || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/nscd || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/pcprofiledump || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/pldd || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/rpcgen || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/sln || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/sotruss || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/sprof || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/tzselect || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/xtrace || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/zdump || true
nios2-linux-gnu-objcopy -R .comment -R .note -R .debug_info -R .debug_aranges -R .debug_pubnames -R .debug_pubtypes -R .debug_abbrev -R .debug_line -R .debug_str -R .debug_ranges -R .debug_loc /scratch/sandra/nios2-linux-bugfix/install/nios2-linux-gnu/libc/.//usr/lib/.//bin/zic || true
popenv

# task [132/219] /i686-pc-linux-gnu/package_tbz2
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
rm -f /scratch/sandra/nios2-linux-bugfix/pkg/sourceryg++-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu.tar.bz2
pushd /scratch/sandra/nios2-linux-bugfix/obj
rm -f sourceryg++-2013.05
ln -s /scratch/sandra/nios2-linux-bugfix/install sourceryg++-2013.05
tar cjf /scratch/sandra/nios2-linux-bugfix/pkg/sourceryg++-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu.tar.bz2 --owner=0 --group=0 --exclude=host-i686-pc-linux-gnu --exclude=host-i686-mingw32 sourceryg++-2013.05/bin sourceryg++-2013.05/i686-pc-linux-gnu sourceryg++-2013.05/lib sourceryg++-2013.05/libexec sourceryg++-2013.05/nios2-linux-gnu sourceryg++-2013.05/share
rm -f sourceryg++-2013.05
popd
popenv

# task [133/219] /i686-pc-linux-gnu/package_installanywhere/unpack
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
/scratch/sandra/nios2-linux-bugfix/src/scripts-trunk/gnu-ia-unpack -i /scratch/sandra/nios2-linux-bugfix/install -l /scratch/sandra/nios2-linux-bugfix/logs -o /scratch/sandra/nios2-linux-bugfix/obj -p /scratch/sandra/nios2-linux-bugfix/pkg -s /scratch/sandra/nios2-linux-bugfix/src -T /scratch/sandra/nios2-linux-bugfix/testlogs -h i686-pc-linux-gnu -e /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras sgxx/2013.05/nios2-linux-gnu-lite
popenv

# task [134/219] /i686-pc-linux-gnu/package_installanywhere/merge_modules
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
/scratch/sandra/nios2-linux-bugfix/src/scripts-trunk/gnu-ia-helper-mm -i /scratch/sandra/nios2-linux-bugfix/install -l /scratch/sandra/nios2-linux-bugfix/logs -o /scratch/sandra/nios2-linux-bugfix/obj -p /scratch/sandra/nios2-linux-bugfix/pkg -s /scratch/sandra/nios2-linux-bugfix/src -T /scratch/sandra/nios2-linux-bugfix/testlogs -h i686-pc-linux-gnu -m iamm-common sgxx/2013.05/nios2-linux-gnu-lite
/scratch/sandra/nios2-linux-bugfix/src/scripts-trunk/gnu-ia-helper-mm -i /scratch/sandra/nios2-linux-bugfix/install -l /scratch/sandra/nios2-linux-bugfix/logs -o /scratch/sandra/nios2-linux-bugfix/obj -p /scratch/sandra/nios2-linux-bugfix/pkg -s /scratch/sandra/nios2-linux-bugfix/src -T /scratch/sandra/nios2-linux-bugfix/testlogs -h i686-pc-linux-gnu -m iamm-jvm sgxx/2013.05/nios2-linux-gnu-lite
popenv

# task [135/219] /i686-pc-linux-gnu/package_installanywhere/installer
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-pc-linux-gnu-gcc '
pushenvvar CXX 'i686-pc-linux-gnu-g++ '
pushenvvar LD 'i686-pc-linux-gnu-ld '
pushenvvar OBJCOPY 'i686-pc-linux-gnu-objcopy '
pushenvvar AR 'i686-pc-linux-gnu-ar '
pushenvvar RANLIB 'i686-pc-linux-gnu-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/install/bin
/scratch/sandra/nios2-linux-bugfix/src/scripts-trunk/gnu-ia-helper -i /scratch/sandra/nios2-linux-bugfix/install -l /scratch/sandra/nios2-linux-bugfix/logs -o /scratch/sandra/nios2-linux-bugfix/obj -p /scratch/sandra/nios2-linux-bugfix/pkg -s /scratch/sandra/nios2-linux-bugfix/src -T /scratch/sandra/nios2-linux-bugfix/testlogs -h i686-pc-linux-gnu -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/getting_started/license.html -e /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras sgxx/2013.05/nios2-linux-gnu-lite
popenv

# task [136/219] /i686-mingw32/host_cleanup
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
popenv

# task [137/219] /i686-mingw32/host_unpack
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32
mkdir /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32
pushd /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32
ln -s . sourceryg++-2013.05
tar xf /scratch/sandra/nios2-linux-bugfix/pkg/sourceryg++-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu.tar.bz2 --bzip2
rm sourceryg++-2013.05
popd
popenv

# task [138/219] /i686-mingw32/libiconv/0/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/libiconv-2013.05-43-nios2-linux-gnu-i686-mingw32
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/libiconv-2013.05-43-nios2-linux-gnu-i686-mingw32
pushd /scratch/sandra/nios2-linux-bugfix/obj/libiconv-2013.05-43-nios2-linux-gnu-i686-mingw32
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/libiconv-1.11 --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --disable-shared --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-mingw32 --disable-nls
popd
popenv
popenv
popenv

# task [139/219] /i686-mingw32/libiconv/0/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/libiconv-2013.05-43-nios2-linux-gnu-i686-mingw32
make -j1
popd
popenv
popenv
popenv

# task [140/219] /i686-mingw32/libiconv/0/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/libiconv-2013.05-43-nios2-linux-gnu-i686-mingw32
make install
popd
popenv
popenv
popenv

# task [141/219] /i686-mingw32/make/copy
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/make-src-2013.05-43-nios2-linux-gnu-i686-mingw32
copy_dir_clean /scratch/sandra/nios2-linux-bugfix/src/make-3.81 /scratch/sandra/nios2-linux-bugfix/obj/make-src-2013.05-43-nios2-linux-gnu-i686-mingw32
chmod -R u+w /scratch/sandra/nios2-linux-bugfix/obj/make-src-2013.05-43-nios2-linux-gnu-i686-mingw32
touch /scratch/sandra/nios2-linux-bugfix/obj/make-src-2013.05-43-nios2-linux-gnu-i686-mingw32/.gnu-stamp
popenv
popenv
popenv

# task [142/219] /i686-mingw32/make/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/make-2013.05-43-nios2-linux-gnu-i686-mingw32
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/make-2013.05-43-nios2-linux-gnu-i686-mingw32
pushd /scratch/sandra/nios2-linux-bugfix/obj/make-2013.05-43-nios2-linux-gnu-i686-mingw32
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/obj/make-src-2013.05-43-nios2-linux-gnu-i686-mingw32 --prefix=/opt/codesourcery --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-mingw32 '--with-pkgversion=Sourcery CodeBench Lite 2013.05-43' --disable-nls --program-prefix=cs-
popd
popenv
popenv
popenv

# task [143/219] /i686-mingw32/make/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/make-2013.05-43-nios2-linux-gnu-i686-mingw32
make -j4
popd
popenv
popenv
popenv

# task [144/219] /i686-mingw32/make/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/make-2013.05-43-nios2-linux-gnu-i686-mingw32
make install prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 exec_prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 libdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/man datadir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share
popd
popenv
popenv
popenv

# task [145/219] /i686-mingw32/coreutils/copy
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/coreutils-src-2013.05-43-nios2-linux-gnu-i686-mingw32
copy_dir_clean /scratch/sandra/nios2-linux-bugfix/src/coreutils-5.94 /scratch/sandra/nios2-linux-bugfix/obj/coreutils-src-2013.05-43-nios2-linux-gnu-i686-mingw32
chmod -R u+w /scratch/sandra/nios2-linux-bugfix/obj/coreutils-src-2013.05-43-nios2-linux-gnu-i686-mingw32
touch /scratch/sandra/nios2-linux-bugfix/obj/coreutils-src-2013.05-43-nios2-linux-gnu-i686-mingw32/.gnu-stamp
popenv
popenv
popenv

# task [146/219] /i686-mingw32/coreutils/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/coreutils-2013.05-43-nios2-linux-gnu-i686-mingw32
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/coreutils-2013.05-43-nios2-linux-gnu-i686-mingw32
pushd /scratch/sandra/nios2-linux-bugfix/obj/coreutils-2013.05-43-nios2-linux-gnu-i686-mingw32
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/obj/coreutils-src-2013.05-43-nios2-linux-gnu-i686-mingw32 --prefix=/opt/codesourcery --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-mingw32 '--with-pkgversion=Sourcery CodeBench Lite 2013.05-43' --disable-nls --program-prefix=cs-
popd
popenv
popenv
popenv

# task [147/219] /i686-mingw32/coreutils/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/coreutils-2013.05-43-nios2-linux-gnu-i686-mingw32
make -j4
popd
popenv
popenv
popenv

# task [148/219] /i686-mingw32/coreutils/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/coreutils-2013.05-43-nios2-linux-gnu-i686-mingw32
make install prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 exec_prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 libdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/man datadir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share
popd
popenv
popenv
popenv

# task [149/219] /i686-mingw32/zlib_first/copy
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/zlib-first-2013.05-43-nios2-linux-gnu-i686-mingw32
copy_dir_clean /scratch/sandra/nios2-linux-bugfix/src/zlib-1.2.7 /scratch/sandra/nios2-linux-bugfix/obj/zlib-first-2013.05-43-nios2-linux-gnu-i686-mingw32
chmod -R u+w /scratch/sandra/nios2-linux-bugfix/obj/zlib-first-2013.05-43-nios2-linux-gnu-i686-mingw32
popenv

# task [150/219] /i686-mingw32/zlib_first/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushd /scratch/sandra/nios2-linux-bugfix/obj/zlib-first-2013.05-43-nios2-linux-gnu-i686-mingw32
pushenv
pushenvvar CFLAGS '-O3 -fPIC'
pushenvvar CC 'i686-mingw32-gcc  '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
./configure --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --static
popenv
popd
popenv

# task [151/219] /i686-mingw32/zlib_first/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushd /scratch/sandra/nios2-linux-bugfix/obj/zlib-first-2013.05-43-nios2-linux-gnu-i686-mingw32
make -j4
popd
popenv

# task [152/219] /i686-mingw32/zlib_first/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushd /scratch/sandra/nios2-linux-bugfix/obj/zlib-first-2013.05-43-nios2-linux-gnu-i686-mingw32
make install
popd
popenv

# task [153/219] /i686-mingw32/gmp/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushenvvar CFLAGS '-g -O2'
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/gmp-2013.05-43-nios2-linux-gnu-i686-mingw32
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/gmp-2013.05-43-nios2-linux-gnu-i686-mingw32
pushd /scratch/sandra/nios2-linux-bugfix/obj/gmp-2013.05-43-nios2-linux-gnu-i686-mingw32
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/gmp-2013.05 --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --disable-shared --build=i686-pc-linux-gnu --target=i686-mingw32 --host=i686-mingw32 --enable-cxx --disable-nls
popd
popenv
popenv
popenv

# task [154/219] /i686-mingw32/gmp/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushenvvar CFLAGS '-g -O2'
pushd /scratch/sandra/nios2-linux-bugfix/obj/gmp-2013.05-43-nios2-linux-gnu-i686-mingw32
make -j4
popd
popenv
popenv
popenv

# task [155/219] /i686-mingw32/gmp/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushenvvar CFLAGS '-g -O2'
pushd /scratch/sandra/nios2-linux-bugfix/obj/gmp-2013.05-43-nios2-linux-gnu-i686-mingw32
make install
popd
popenv
popenv
popenv

# task [156/219] /i686-mingw32/gmp/postinstall
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushenvvar CFLAGS '-g -O2'
popenv
popenv
popenv

# task [157/219] /i686-mingw32/mpfr/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/mpfr-2013.05-43-nios2-linux-gnu-i686-mingw32
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/mpfr-2013.05-43-nios2-linux-gnu-i686-mingw32
pushd /scratch/sandra/nios2-linux-bugfix/obj/mpfr-2013.05-43-nios2-linux-gnu-i686-mingw32
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/mpfr-2013.05 --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --disable-shared --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-mingw32 --disable-nls --with-gmp=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr
popd
popenv
popenv
popenv

# task [158/219] /i686-mingw32/mpfr/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/mpfr-2013.05-43-nios2-linux-gnu-i686-mingw32
make -j4
popd
popenv
popenv
popenv

# task [159/219] /i686-mingw32/mpfr/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/mpfr-2013.05-43-nios2-linux-gnu-i686-mingw32
make install
popd
popenv
popenv
popenv

# task [160/219] /i686-mingw32/mpfr/postinstall
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
popenv
popenv
popenv

# task [161/219] /i686-mingw32/mpc/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/mpc-2013.05-43-nios2-linux-gnu-i686-mingw32
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/mpc-2013.05-43-nios2-linux-gnu-i686-mingw32
pushd /scratch/sandra/nios2-linux-bugfix/obj/mpc-2013.05-43-nios2-linux-gnu-i686-mingw32
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/mpc-2013.05 --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --disable-shared --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-mingw32 --disable-nls --with-gmp=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --with-mpfr=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr
popd
popenv
popenv
popenv

# task [162/219] /i686-mingw32/mpc/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/mpc-2013.05-43-nios2-linux-gnu-i686-mingw32
make -j4
popd
popenv
popenv
popenv

# task [163/219] /i686-mingw32/mpc/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/mpc-2013.05-43-nios2-linux-gnu-i686-mingw32
make install
popd
popenv
popenv
popenv

# task [164/219] /i686-mingw32/mpc/postinstall
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
popenv
popenv
popenv

# task [165/219] /i686-mingw32/ppl/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/ppl-2013.05-43-nios2-linux-gnu-i686-mingw32
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/ppl-2013.05-43-nios2-linux-gnu-i686-mingw32
pushd /scratch/sandra/nios2-linux-bugfix/obj/ppl-2013.05-43-nios2-linux-gnu-i686-mingw32
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/ppl-sgxx-0.11-2013.05 --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --disable-shared --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-mingw32 --disable-nls --with-libgmp-prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --with-gmp-prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr CPPFLAGS=-I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/include LDFLAGS=-L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/lib --disable-watchdog
popd
popenv
popenv
popenv

# task [166/219] /i686-mingw32/ppl/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/ppl-2013.05-43-nios2-linux-gnu-i686-mingw32
make -j4
popd
popenv
popenv
popenv

# task [167/219] /i686-mingw32/ppl/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/ppl-2013.05-43-nios2-linux-gnu-i686-mingw32
make install
popd
popenv
popenv
popenv

# task [168/219] /i686-mingw32/cloog/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/cloog-2013.05-43-nios2-linux-gnu-i686-mingw32
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/cloog-2013.05-43-nios2-linux-gnu-i686-mingw32
pushd /scratch/sandra/nios2-linux-bugfix/obj/cloog-2013.05-43-nios2-linux-gnu-i686-mingw32
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/cloog-0.15 --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --disable-shared --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-mingw32 --disable-nls --with-ppl=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --with-gmp=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr
popd
popenv
popenv
popenv

# task [169/219] /i686-mingw32/cloog/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/cloog-2013.05-43-nios2-linux-gnu-i686-mingw32
make -j4
popd
popenv
popenv
popenv

# task [170/219] /i686-mingw32/cloog/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/cloog-2013.05-43-nios2-linux-gnu-i686-mingw32
make install
popd
popenv
popenv
popenv

# task [171/219] /i686-mingw32/cloog/postinstall
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
popenv
popenv
popenv

# task [172/219] /i686-mingw32/libelf/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/libelf-2013.05-43-nios2-linux-gnu-i686-mingw32
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/libelf-2013.05-43-nios2-linux-gnu-i686-mingw32
pushd /scratch/sandra/nios2-linux-bugfix/obj/libelf-2013.05-43-nios2-linux-gnu-i686-mingw32
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/libelf-2013.05 --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --disable-shared --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-mingw32 --disable-nls
popd
popenv
popenv
popenv

# task [173/219] /i686-mingw32/libelf/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/libelf-2013.05-43-nios2-linux-gnu-i686-mingw32
make -j4
popd
popenv
popenv
popenv

# task [174/219] /i686-mingw32/libelf/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/libelf-2013.05-43-nios2-linux-gnu-i686-mingw32
make install
popd
popenv
popenv
popenv

# task [175/219] /i686-mingw32/toolchain/binutils/copy
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/lib
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/binutils-src-2013.05-43-nios2-linux-gnu-i686-mingw32
copy_dir_clean /scratch/sandra/nios2-linux-bugfix/src/binutils-2013.05 /scratch/sandra/nios2-linux-bugfix/obj/binutils-src-2013.05-43-nios2-linux-gnu-i686-mingw32
chmod -R u+w /scratch/sandra/nios2-linux-bugfix/obj/binutils-src-2013.05-43-nios2-linux-gnu-i686-mingw32
touch /scratch/sandra/nios2-linux-bugfix/obj/binutils-src-2013.05-43-nios2-linux-gnu-i686-mingw32/.gnu-stamp
popenv
popenv
popenv

# task [176/219] /i686-mingw32/toolchain/binutils/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/lib
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-mingw32
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-mingw32
pushd /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-mingw32
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/obj/binutils-src-2013.05-43-nios2-linux-gnu-i686-mingw32 --prefix=/opt/codesourcery --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-mingw32 '--with-pkgversion=Sourcery CodeBench Lite 2013.05-43' --with-bugurl=https://sourcery.mentor.com/GNUToolchain/ --disable-nls --with-sysroot=/opt/codesourcery/nios2-linux-gnu/libc --enable-poison-system-directories --enable-plugins
popd
popenv
popenv
popenv

# task [177/219] /i686-mingw32/toolchain/binutils/libiberty
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/lib
pushd /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-mingw32
make -j4 all-libiberty
popd
copy_dir_clean /scratch/sandra/nios2-linux-bugfix/src/binutils-2013.05/include /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-mingw32/usr/include
chmod -R u+w /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-mingw32/usr/include
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-mingw32/usr/lib
cp /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-mingw32/libiberty/libiberty.a /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-mingw32/usr/lib
popenv
popenv
popenv

# task [178/219] /i686-mingw32/toolchain/binutils/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/lib
pushd /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-mingw32
make -j4
popd
popenv
popenv
popenv

# task [179/219] /i686-mingw32/toolchain/binutils/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/lib
pushd /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-mingw32
make install prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 exec_prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 libdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/man datadir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share
popd
popenv
popenv
popenv

# task [180/219] /i686-mingw32/toolchain/binutils/postinstall
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/lib
pushd /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32
rm lib/charset.alias
rm ./lib/libiberty.a
rmdir ./lib
popd
pushd /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-mingw32
make prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 exec_prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 libdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/man datadir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share install-html
popd
pushd /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-mingw32
make prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 exec_prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 libdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/man datadir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share install-pdf
popd
cp /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-mingw32/bfd/.libs/libbfd.a /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-mingw32/usr/lib
cp /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-mingw32/bfd/bfd.h /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-mingw32/usr/include
cp /scratch/sandra/nios2-linux-bugfix/src/binutils-2013.05/bfd/elf-bfd.h /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-mingw32/usr/include
cp /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-mingw32/opcodes/.libs/libopcodes.a /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-mingw32/usr/lib
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-mingw32/usr/testbin
cp /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-mingw32/binutils/bfdtest1.exe /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-mingw32/usr/testbin
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-mingw32/usr/testbin
cp /scratch/sandra/nios2-linux-bugfix/obj/binutils-2013.05-43-nios2-linux-gnu-i686-mingw32/binutils/bfdtest2.exe /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-binutils-i686-mingw32/usr/testbin
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html/configure.html /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf/etc/configure.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info/configure.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly configure
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html/standards.html /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf/etc/standards.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info/standards.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly standards
rmdir /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf/etc
rm -rf /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html/bfd.html /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf/bfd.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info/bfd.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly bfd
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html/libiberty.html /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf/libiberty.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-ld.bfd.exe
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/ld.bfd.exe
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/bin/ld.bfd.exe
popenv
popenv
popenv

# task [181/219] /i686-mingw32/toolchain/copy_libs
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
rm -rf /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc
copy_dir /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/nios2-linux-gnu/lib /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/lib
copy_dir /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/nios2-linux-gnu/libc /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc
copy_dir /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/nios2-linux-gnu/include/c++ /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/include/c++
copy_dir /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/lib/gcc/nios2-linux-gnu /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/lib/gcc/nios2-linux-gnu
copy_dir /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html
copy_dir /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf
copy_dir /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info
copy_dir /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/man /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/man
cp /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/LICENSE.txt /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu
popenv

# task [182/219] /i686-mingw32/toolchain/gcc_final/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenvvar AR_FOR_TARGET nios2-linux-gnu-ar
pushenvvar NM_FOR_TARGET nios2-linux-gnu-nm
pushenvvar OBJDUMP_FOR_TARET nios2-linux-gnu-objdump
pushenvvar STRIP_FOR_TARGET nios2-linux-gnu-strip
pushenvvar CC_FOR_TARGET nios2-linux-gnu-gcc
pushenvvar GCC_FOR_TARGET nios2-linux-gnu-gcc
pushenvvar CXX_FOR_TARGET nios2-linux-gnu-g++
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/gcc-2013.05-43-nios2-linux-gnu-i686-mingw32
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/gcc-2013.05-43-nios2-linux-gnu-i686-mingw32
pushd /scratch/sandra/nios2-linux-bugfix/obj/gcc-2013.05-43-nios2-linux-gnu-i686-mingw32
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/gcc-nios2 --build=i686-pc-linux-gnu --host=i686-mingw32 --target=nios2-linux-gnu --enable-threads --disable-libmudflap --disable-libssp --disable-libstdcxx-pch --with-gnu-as --with-gnu-ld '--with-specs=%{save-temps: -fverbose-asm} -D__CS_SOURCERYGXX_MAJ__=2013 -D__CS_SOURCERYGXX_MIN__=5 -D__CS_SOURCERYGXX_REV__=43 %{O2:%{!fno-remove-local-statics: -fremove-local-statics}} %{O*:%{O|O0|O1|O2|Os:;:%{!fno-remove-local-statics: -fremove-local-statics}}}' --enable-languages=c,c++ --enable-shared --enable-lto --enable-symvers=gnu --enable-__cxa_atexit '--with-pkgversion=Sourcery CodeBench Lite 2013.05-43' --with-bugurl=https://sourcery.mentor.com/GNUToolchain/ --disable-nls --prefix=/opt/codesourcery --with-sysroot=/opt/codesourcery/nios2-linux-gnu/libc --with-build-sysroot=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc --with-libiconv-prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --with-gmp=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --with-mpfr=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --with-mpc=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --with-ppl=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr '--with-host-libstdcxx=-static-libgcc -Wl,-Bstatic,-lstdc++,-Bdynamic -lm' --with-cloog=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --with-libelf=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --disable-libgomp --disable-libitm --enable-poison-system-directories --with-build-time-tools=/scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/nios2-linux-gnu/bin --with-build-time-tools=/scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/nios2-linux-gnu/bin SED=sed
popd
popenv
popenv

# task [183/219] /i686-mingw32/toolchain/gcc_final/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenvvar AR_FOR_TARGET nios2-linux-gnu-ar
pushenvvar NM_FOR_TARGET nios2-linux-gnu-nm
pushenvvar OBJDUMP_FOR_TARET nios2-linux-gnu-objdump
pushenvvar STRIP_FOR_TARGET nios2-linux-gnu-strip
pushenvvar CC_FOR_TARGET nios2-linux-gnu-gcc
pushenvvar GCC_FOR_TARGET nios2-linux-gnu-gcc
pushenvvar CXX_FOR_TARGET nios2-linux-gnu-g++
pushd /scratch/sandra/nios2-linux-bugfix/obj/gcc-2013.05-43-nios2-linux-gnu-i686-mingw32
make -j4 LDFLAGS_FOR_TARGET=--sysroot=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc CPPFLAGS_FOR_TARGET=--sysroot=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc build_tooldir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu all-gcc
popd
popenv
popenv

# task [184/219] /i686-mingw32/toolchain/gcc_final/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenvvar AR_FOR_TARGET nios2-linux-gnu-ar
pushenvvar NM_FOR_TARGET nios2-linux-gnu-nm
pushenvvar OBJDUMP_FOR_TARET nios2-linux-gnu-objdump
pushenvvar STRIP_FOR_TARGET nios2-linux-gnu-strip
pushenvvar CC_FOR_TARGET nios2-linux-gnu-gcc
pushenvvar GCC_FOR_TARGET nios2-linux-gnu-gcc
pushenvvar CXX_FOR_TARGET nios2-linux-gnu-g++
pushd /scratch/sandra/nios2-linux-bugfix/obj/gcc-2013.05-43-nios2-linux-gnu-i686-mingw32
make prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 exec_prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 libdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/man install-gcc
make prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 exec_prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 libdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/man install-html-gcc
make prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 exec_prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 libdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/man install-pdf-gcc
popd
popenv
popenv

# task [185/219] /i686-mingw32/toolchain/gcc_final/postinstall
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenvvar AR_FOR_TARGET nios2-linux-gnu-ar
pushenvvar NM_FOR_TARGET nios2-linux-gnu-nm
pushenvvar OBJDUMP_FOR_TARET nios2-linux-gnu-objdump
pushenvvar STRIP_FOR_TARGET nios2-linux-gnu-strip
pushenvvar CC_FOR_TARGET nios2-linux-gnu-gcc
pushenvvar GCC_FOR_TARGET nios2-linux-gnu-gcc
pushenvvar CXX_FOR_TARGET nios2-linux-gnu-g++
pushd /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32
rmdir include
popd
copy_dir /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/lib/gcc/nios2-linux-gnu /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/lib/gcc/nios2-linux-gnu
copy_dir /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/nios2-linux-gnu/lib /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/lib
copy_dir /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/nios2-linux-gnu/include/c++ /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/include/c++
rm -rf /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html/gccinstall /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf/gcc/gccinstall.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info/gccinstall.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly gccinstall
rm -rf /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html/gccint /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf/gcc/gccint.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info/gccint.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly gccint
rm -rf /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html/cppinternals /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf/gcc/cppinternals.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info/cppinternals.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly cppinternals
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html/libiberty.html /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf/libiberty.pdf
popenv
popenv

# task [186/219] /i686-mingw32/toolchain/zlib/0/copy
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/zlib-2013.05-43-nios2-linux-gnu-i686-mingw32
copy_dir_clean /scratch/sandra/nios2-linux-bugfix/src/zlib-1.2.7 /scratch/sandra/nios2-linux-bugfix/obj/zlib-2013.05-43-nios2-linux-gnu-i686-mingw32
chmod -R u+w /scratch/sandra/nios2-linux-bugfix/obj/zlib-2013.05-43-nios2-linux-gnu-i686-mingw32
popenv

# task [187/219] /i686-mingw32/toolchain/zlib/0/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushd /scratch/sandra/nios2-linux-bugfix/obj/zlib-2013.05-43-nios2-linux-gnu-i686-mingw32
pushenv
pushenvvar CFLAGS '-O3 -fPIC'
pushenvvar CC 'i686-mingw32-gcc  '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
./configure --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --static
popenv
popd
popenv

# task [188/219] /i686-mingw32/toolchain/zlib/0/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushd /scratch/sandra/nios2-linux-bugfix/obj/zlib-2013.05-43-nios2-linux-gnu-i686-mingw32
make -j4
popd
popenv

# task [189/219] /i686-mingw32/toolchain/zlib/0/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushd /scratch/sandra/nios2-linux-bugfix/obj/zlib-2013.05-43-nios2-linux-gnu-i686-mingw32
make install
popd
popenv

# task [190/219] /i686-mingw32/toolchain/expat/0/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/expat-2013.05-43-nios2-linux-gnu-i686-mingw32
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/expat-2013.05-43-nios2-linux-gnu-i686-mingw32
pushd /scratch/sandra/nios2-linux-bugfix/obj/expat-2013.05-43-nios2-linux-gnu-i686-mingw32
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/expat-2013.05 --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --disable-shared --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-mingw32 --disable-nls
popd
popenv
popenv
popenv

# task [191/219] /i686-mingw32/toolchain/expat/0/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/expat-2013.05-43-nios2-linux-gnu-i686-mingw32
make -j4
popd
popenv
popenv
popenv

# task [192/219] /i686-mingw32/toolchain/expat/0/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/expat-2013.05-43-nios2-linux-gnu-i686-mingw32
make install
popd
popenv
popenv
popenv

# task [193/219] /i686-mingw32/toolchain/gdb/0/copy
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/lib
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/gdb-src-2013.05-43-nios2-linux-gnu-i686-mingw32
copy_dir_clean /scratch/sandra/nios2-linux-bugfix/src/gdb-nios2 /scratch/sandra/nios2-linux-bugfix/obj/gdb-src-2013.05-43-nios2-linux-gnu-i686-mingw32
chmod -R u+w /scratch/sandra/nios2-linux-bugfix/obj/gdb-src-2013.05-43-nios2-linux-gnu-i686-mingw32
touch /scratch/sandra/nios2-linux-bugfix/obj/gdb-src-2013.05-43-nios2-linux-gnu-i686-mingw32/.gnu-stamp
popenv
popenv
popenv

# task [194/219] /i686-mingw32/toolchain/gdb/0/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/lib
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/gdb-2013.05-43-nios2-linux-gnu-i686-mingw32
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/gdb-2013.05-43-nios2-linux-gnu-i686-mingw32
pushd /scratch/sandra/nios2-linux-bugfix/obj/gdb-2013.05-43-nios2-linux-gnu-i686-mingw32
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/obj/gdb-src-2013.05-43-nios2-linux-gnu-i686-mingw32 --prefix=/opt/codesourcery --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-mingw32 '--with-pkgversion=Sourcery CodeBench Lite 2013.05-43' --with-bugurl=https://sourcery.mentor.com/GNUToolchain/ --disable-libmcheck --disable-nls --with-libexpat-prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --with-libiconv-prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --with-system-gdbinit=/opt/codesourcery/i686-mingw32/nios2-linux-gnu/lib/gdbinit '--with-gdb-datadir='\''${prefix}'\''/nios2-linux-gnu/share/gdb'
popd
popenv
popenv
popenv

# task [195/219] /i686-mingw32/toolchain/gdb/0/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/lib
pushd /scratch/sandra/nios2-linux-bugfix/obj/gdb-2013.05-43-nios2-linux-gnu-i686-mingw32
make -j4
popd
popenv
popenv
popenv

# task [196/219] /i686-mingw32/toolchain/gdb/0/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/lib
pushd /scratch/sandra/nios2-linux-bugfix/obj/gdb-2013.05-43-nios2-linux-gnu-i686-mingw32
make install prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 exec_prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 libdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/man datadir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share
popd
popenv
popenv
popenv

# task [197/219] /i686-mingw32/toolchain/gdb/0/postinstall
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushenvvar CPPFLAGS -I/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/include
pushenvvar LDFLAGS -L/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr/lib
pushd /scratch/sandra/nios2-linux-bugfix/obj/gdb-2013.05-43-nios2-linux-gnu-i686-mingw32
make prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 exec_prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 libdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/man datadir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share install-html
make prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 exec_prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 libdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/man datadir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share install-pdf
popd
rm -rf /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html/annotate /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf/annotate.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info/annotate.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly annotate
rm -rf /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html/gdbint /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf/gdbint.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info/gdbint.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly gdbint
rm -rf /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html/stabs /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf/stabs.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info/stabs.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly stabs
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html/configure.html /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf/etc/configure.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info/configure.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly configure
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html/standards.html /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf/etc/standards.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info/standards.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly standards
rmdir /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf/etc
rm -rf /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html/bfd.html /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf/bfd.pdf
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info/bfd.info
install-info --infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info --remove-exactly bfd
rm -f /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html/libiberty.html /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf/libiberty.pdf
pushd /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32
mkdir -p i686-mingw32/nios2-linux-gnu/include/
mv include/gdb i686-mingw32/nios2-linux-gnu/include
rmdir include
popd
popenv
popenv
popenv

# task [198/219] /i686-mingw32/gdil/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/gdil-2013.05-43-nios2-linux-gnu-i686-mingw32
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/gdil-2013.05-43-nios2-linux-gnu-i686-mingw32
pushd /scratch/sandra/nios2-linux-bugfix/obj/gdil-2013.05-43-nios2-linux-gnu-i686-mingw32
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/gdil-2013.05 --prefix=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --disable-shared --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-mingw32 --with-csl-docbook=/scratch/sandra/nios2-linux-bugfix/src/csl-docbook-trunk '--with-xml-catalog-files=/usr/local/tools/gcc-4.7.3/share/sgml/docbook/docbook-xsl/catalog.xml /etc/xml/catalog' --disable-nls --with-expat=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr '--with-features=xml autoconf driver config sprite doc'
popd
popenv
popenv
popenv

# task [199/219] /i686-mingw32/gdil/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/gdil-2013.05-43-nios2-linux-gnu-i686-mingw32
make -j4
popd
popenv
popenv
popenv

# task [200/219] /i686-mingw32/gdil/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/gdil-2013.05-43-nios2-linux-gnu-i686-mingw32
make install
popd
popenv
popenv
popenv

# task [201/219] /i686-mingw32/edge_utils/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/edge-utils-2013.05-43-nios2-linux-gnu-i686-mingw32
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/edge-utils-2013.05-43-nios2-linux-gnu-i686-mingw32
copy_dir_clean /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source /scratch/sandra/nios2-linux-bugfix/obj/edge-utils-2013.05-43-nios2-linux-gnu-i686-mingw32/source
chmod -R u+w /scratch/sandra/nios2-linux-bugfix/obj/edge-utils-2013.05-43-nios2-linux-gnu-i686-mingw32
make -C /scratch/sandra/nios2-linux-bugfix/obj/edge-utils-2013.05-43-nios2-linux-gnu-i686-mingw32/source/debugger/utils -j4 BLDTYPE=rls HOST_OS=win32 GPREF=i686-mingw32- win32_rls/libUtils.a
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/edge-utils-i686-mingw32/lib
cp /scratch/sandra/nios2-linux-bugfix/obj/edge-utils-2013.05-43-nios2-linux-gnu-i686-mingw32/source/debugger/utils/win32_rls/libUtils.a /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/edge-utils-i686-mingw32/lib
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/edge-utils-i686-mingw32/include
cp /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/NumericValue.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/clBaseTypes.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/clByteBuffer.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/clErrorLogger.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/clException.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/clStl.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluAllocators.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluAtomic.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluCLibrary.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluCond.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluConfigData.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluConfigFile.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluDebugTrace.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluDll.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluEmo.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluEndian.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluErrors.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluEvent.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluEventWaiter.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluFile.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluFileUtils.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluFindFile.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluFunctionTrace.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluGuard.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluHighResTimer.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluHtmlLogFile.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluInetPackets.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluInetSocket.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluLogFile.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluMutex.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluOutput.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluQ.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluRedirect.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluRegex.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluSemaphore.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluSerialPort.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluSerialPortBase.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluSerialPortPosix.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluSerialPortWin32.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluSmartPtr.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluSocketBase.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluSplitPath.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluStats.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluStringHeapManager.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluStringUtil.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluSync.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluTask.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluTextTable.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluThread.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluThreadPool.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluThreadedMember.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluTime.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluTimerQueue.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluTransaction.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluTransactionIdMap.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluUniqueIDManager.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluUnixSignalHandling.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluUtils.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluUtilsPch.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluValueMap.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluVector.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluXmlParser.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/rpn_calc.h /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/stdint_opus.h /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/edge-utils-i686-mingw32/include
cp /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluConfigData.inl /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluEvent.inl /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluGuard.inl /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluHighResTimer.inl /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluLogFile.inl /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluMutex.inl /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluTask.inl /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluThread.inl /scratch/sandra/nios2-linux-bugfix/src/edge-utils-2013.05/source/debugger/utils/cluTime.inl /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/edge-utils-i686-mingw32/include
popenv

# task [202/219] /i686-mingw32/sprite/configure
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
rm -rf /scratch/sandra/nios2-linux-bugfix/obj/sprite-2013.05-43-nios2-linux-gnu-i686-mingw32
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/sprite-2013.05-43-nios2-linux-gnu-i686-mingw32
pushd /scratch/sandra/nios2-linux-bugfix/obj/sprite-2013.05-43-nios2-linux-gnu-i686-mingw32
check_mentor_trademarks /scratch/sandra/nios2-linux-bugfix/src/sprite-2013.05 --prefix=/opt/codesourcery --build=i686-pc-linux-gnu --target=nios2-linux-gnu --host=i686-mingw32 '--with-pkgversion=Sourcery CodeBench Lite 2013.05-43' --with-bugurl=https://sourcery.mentor.com/GNUToolchain/ --enable-backends=usbblaster --disable-nls --with-gdil=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --with-expat=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/host-libs-i686-mingw32/usr --with-boards=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/cs3/boards --with-edge-utils=/scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/edge-utils-i686-mingw32 --with-mep_cs=
popd
popenv
popenv
popenv

# task [203/219] /i686-mingw32/sprite/build
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/sprite-2013.05-43-nios2-linux-gnu-i686-mingw32
make -j4
popd
popenv
popenv
popenv

# task [204/219] /i686-mingw32/sprite/install
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
pushd /scratch/sandra/nios2-linux-bugfix/obj/sprite-2013.05-43-nios2-linux-gnu-i686-mingw32
make install prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 exec_prefix=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 libdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/lib htmldir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/html pdfdir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/pdf infodir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/info mandir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share/doc/sourceryg++-nios2-linux-gnu/man datadir=/scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/share
popd
popenv
popenv
popenv

# task [205/219] /i686-mingw32/sprite/postinstall
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushenv
pushenv
popenv
popenv
popenv

# task [206/219] /i686-mingw32/pretidy_installation
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushd /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32
rm ./lib/libiberty.a
popd
popenv

# task [207/219] /i686-mingw32/remove_libtool_archives
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
find /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 -name '*.la' -exec rm '{}' ';'
popenv

# task [208/219] /i686-mingw32/remove_copied_libs
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
rm -f
popenv

# task [209/219] /i686-mingw32/remove_fixed_headers
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
pushd /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/lib/gcc/nios2-linux-gnu/4.7.3/include-fixed
popd
popenv

# task [210/219] /i686-mingw32/add_tooldir_readme
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
cat > /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/bin/README.txt <<'EOF0'
The executables in this directory are for internal use by the compiler
and may not operate correctly when used directly.  This directory
should not be placed on your PATH.  Instead, you should use the
executables in ../../bin/ and place that directory on your PATH.
EOF0
popenv

# task [211/219] /i686-mingw32/strip_host_objects
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/cs-make.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/cs-rm.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-addr2line.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-ar.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-as.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-c++.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-c++filt.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-cpp.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-elfedit.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-g++.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-gcc-4.7.3.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-gcc-ar.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-gcc-nm.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-gcc-ranlib.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-gcc.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-gcov.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-gdb.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-gprof.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-ld.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-nm.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-objcopy.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-objdump.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-ranlib.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-readelf.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-size.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-sprite.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-strings.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/bin/nios2-linux-gnu-strip.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/bin/ar.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/bin/as.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/bin/c++.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/bin/g++.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/bin/gcc.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/bin/ld.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/bin/nm.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/bin/objcopy.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/bin/objdump.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/bin/ranlib.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/bin/strip.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/libexec/gcc/nios2-linux-gnu/4.7.3/cc1.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/libexec/gcc/nios2-linux-gnu/4.7.3/collect2.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/libexec/gcc/nios2-linux-gnu/4.7.3/install-tools/fixincl.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/libexec/gcc/nios2-linux-gnu/4.7.3/cc1plus.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/libexec/gcc/nios2-linux-gnu/4.7.3/lto-wrapper.exe
i686-mingw32-strip /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/libexec/gcc/nios2-linux-gnu/4.7.3/lto1.exe
popenv

# task [212/219] /i686-mingw32/package_tbz2
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
rm -f /scratch/sandra/nios2-linux-bugfix/pkg/sourceryg++-2013.05-43-nios2-linux-gnu-i686-mingw32.tar.bz2
pushd /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32
rm ./nios2-linux-gnu/libc/usr/lib/libnss_compat.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libnss_compat-2.17.so ./nios2-linux-gnu/libc/usr/lib/libnss_compat.so
rm ./nios2-linux-gnu/libc/usr/lib/libstdc++.so.6
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/usr/lib/libstdc++.so.6.0.17 ./nios2-linux-gnu/libc/usr/lib/libstdc++.so.6
rm ./nios2-linux-gnu/libc/usr/lib/librt.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/librt-2.17.so ./nios2-linux-gnu/libc/usr/lib/librt.so
rm ./nios2-linux-gnu/libc/usr/lib/libnss_db.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libnss_db-2.17.so ./nios2-linux-gnu/libc/usr/lib/libnss_db.so
rm ./nios2-linux-gnu/libc/usr/lib/libstdc++.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/usr/lib/libstdc++.so.6.0.17 ./nios2-linux-gnu/libc/usr/lib/libstdc++.so
rm ./nios2-linux-gnu/libc/usr/lib/libutil.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libutil-2.17.so ./nios2-linux-gnu/libc/usr/lib/libutil.so
rm ./nios2-linux-gnu/libc/usr/lib/libBrokenLocale.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libBrokenLocale-2.17.so ./nios2-linux-gnu/libc/usr/lib/libBrokenLocale.so
rm ./nios2-linux-gnu/libc/usr/lib/libresolv.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libresolv-2.17.so ./nios2-linux-gnu/libc/usr/lib/libresolv.so
rm ./nios2-linux-gnu/libc/usr/lib/libanl.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libanl-2.17.so ./nios2-linux-gnu/libc/usr/lib/libanl.so
rm ./nios2-linux-gnu/libc/usr/lib/libnsl.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libnsl-2.17.so ./nios2-linux-gnu/libc/usr/lib/libnsl.so
rm ./nios2-linux-gnu/libc/usr/lib/libdl.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libdl-2.17.so ./nios2-linux-gnu/libc/usr/lib/libdl.so
rm ./nios2-linux-gnu/libc/usr/lib/libthread_db.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libthread_db-1.0.so ./nios2-linux-gnu/libc/usr/lib/libthread_db.so
rm ./nios2-linux-gnu/libc/usr/lib/libnss_files.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libnss_files-2.17.so ./nios2-linux-gnu/libc/usr/lib/libnss_files.so
rm ./nios2-linux-gnu/libc/usr/lib/libcidn.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libcidn-2.17.so ./nios2-linux-gnu/libc/usr/lib/libcidn.so
rm ./nios2-linux-gnu/libc/usr/lib/libnss_nis.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libnss_nis-2.17.so ./nios2-linux-gnu/libc/usr/lib/libnss_nis.so
rm ./nios2-linux-gnu/libc/usr/lib/libnss_nisplus.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libnss_nisplus-2.17.so ./nios2-linux-gnu/libc/usr/lib/libnss_nisplus.so
rm ./nios2-linux-gnu/libc/usr/lib/libnss_hesiod.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libnss_hesiod-2.17.so ./nios2-linux-gnu/libc/usr/lib/libnss_hesiod.so
rm ./nios2-linux-gnu/libc/usr/lib/libnss_dns.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libnss_dns-2.17.so ./nios2-linux-gnu/libc/usr/lib/libnss_dns.so
rm ./nios2-linux-gnu/libc/usr/lib/libcrypt.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libcrypt-2.17.so ./nios2-linux-gnu/libc/usr/lib/libcrypt.so
rm ./nios2-linux-gnu/libc/usr/lib/libm.so
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libm-2.17.so ./nios2-linux-gnu/libc/usr/lib/libm.so
rm ./nios2-linux-gnu/libc/lib/libnss_dns.so.2
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libnss_dns-2.17.so ./nios2-linux-gnu/libc/lib/libnss_dns.so.2
rm ./nios2-linux-gnu/libc/lib/libnss_db.so.2
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libnss_db-2.17.so ./nios2-linux-gnu/libc/lib/libnss_db.so.2
rm ./nios2-linux-gnu/libc/lib/libnss_hesiod.so.2
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libnss_hesiod-2.17.so ./nios2-linux-gnu/libc/lib/libnss_hesiod.so.2
rm ./nios2-linux-gnu/libc/lib/libcrypt.so.1
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libcrypt-2.17.so ./nios2-linux-gnu/libc/lib/libcrypt.so.1
rm ./nios2-linux-gnu/libc/lib/libm.so.6
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libm-2.17.so ./nios2-linux-gnu/libc/lib/libm.so.6
rm ./nios2-linux-gnu/libc/lib/libthread_db.so.1
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libthread_db-1.0.so ./nios2-linux-gnu/libc/lib/libthread_db.so.1
rm ./nios2-linux-gnu/libc/lib/libutil.so.1
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libutil-2.17.so ./nios2-linux-gnu/libc/lib/libutil.so.1
rm ./nios2-linux-gnu/libc/lib/libnsl.so.1
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libnsl-2.17.so ./nios2-linux-gnu/libc/lib/libnsl.so.1
rm ./nios2-linux-gnu/libc/lib/ld.so.1
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/ld-2.17.so ./nios2-linux-gnu/libc/lib/ld.so.1
rm ./nios2-linux-gnu/libc/lib/libcidn.so.1
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libcidn-2.17.so ./nios2-linux-gnu/libc/lib/libcidn.so.1
rm ./nios2-linux-gnu/libc/lib/libc.so.6
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libc-2.17.so ./nios2-linux-gnu/libc/lib/libc.so.6
rm ./nios2-linux-gnu/libc/lib/librt.so.1
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/librt-2.17.so ./nios2-linux-gnu/libc/lib/librt.so.1
rm ./nios2-linux-gnu/libc/lib/libpthread.so.0
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libpthread-2.17.so ./nios2-linux-gnu/libc/lib/libpthread.so.0
rm ./nios2-linux-gnu/libc/lib/libdl.so.2
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libdl-2.17.so ./nios2-linux-gnu/libc/lib/libdl.so.2
rm ./nios2-linux-gnu/libc/lib/libnss_files.so.2
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libnss_files-2.17.so ./nios2-linux-gnu/libc/lib/libnss_files.so.2
rm ./nios2-linux-gnu/libc/lib/libnss_compat.so.2
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libnss_compat-2.17.so ./nios2-linux-gnu/libc/lib/libnss_compat.so.2
rm ./nios2-linux-gnu/libc/lib/libBrokenLocale.so.1
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libBrokenLocale-2.17.so ./nios2-linux-gnu/libc/lib/libBrokenLocale.so.1
rm ./nios2-linux-gnu/libc/lib/libnss_nis.so.2
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libnss_nis-2.17.so ./nios2-linux-gnu/libc/lib/libnss_nis.so.2
rm ./nios2-linux-gnu/libc/lib/libanl.so.1
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libanl-2.17.so ./nios2-linux-gnu/libc/lib/libanl.so.1
rm ./nios2-linux-gnu/libc/lib/libresolv.so.2
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libresolv-2.17.so ./nios2-linux-gnu/libc/lib/libresolv.so.2
rm ./nios2-linux-gnu/libc/lib/libnss_nisplus.so.2
ln /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32/nios2-linux-gnu/libc/lib/libnss_nisplus-2.17.so ./nios2-linux-gnu/libc/lib/libnss_nisplus.so.2
popd
pushd /scratch/sandra/nios2-linux-bugfix/obj
rm -f sourceryg++-2013.05
ln -s /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 sourceryg++-2013.05
tar cjf /scratch/sandra/nios2-linux-bugfix/pkg/sourceryg++-2013.05-43-nios2-linux-gnu-i686-mingw32.tar.bz2 --owner=0 --group=0 --exclude=host-i686-pc-linux-gnu --exclude=host-i686-mingw32 sourceryg++-2013.05/bin sourceryg++-2013.05/i686-mingw32 sourceryg++-2013.05/lib sourceryg++-2013.05/libexec sourceryg++-2013.05/nios2-linux-gnu sourceryg++-2013.05/share
rm -f sourceryg++-2013.05
popd
popenv

# task [213/219] /i686-mingw32/package_installanywhere/unpack
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
/scratch/sandra/nios2-linux-bugfix/src/scripts-trunk/gnu-ia-unpack -i /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 -l /scratch/sandra/nios2-linux-bugfix/logs -o /scratch/sandra/nios2-linux-bugfix/obj -p /scratch/sandra/nios2-linux-bugfix/pkg -s /scratch/sandra/nios2-linux-bugfix/src -T /scratch/sandra/nios2-linux-bugfix/testlogs -h i686-mingw32 -e /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras sgxx/2013.05/nios2-linux-gnu-lite
popenv

# task [214/219] /i686-mingw32/package_installanywhere/merge_modules
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
/scratch/sandra/nios2-linux-bugfix/src/scripts-trunk/gnu-ia-helper-mm -i /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 -l /scratch/sandra/nios2-linux-bugfix/logs -o /scratch/sandra/nios2-linux-bugfix/obj -p /scratch/sandra/nios2-linux-bugfix/pkg -s /scratch/sandra/nios2-linux-bugfix/src -T /scratch/sandra/nios2-linux-bugfix/testlogs -h i686-mingw32 -m iamm-common sgxx/2013.05/nios2-linux-gnu-lite
/scratch/sandra/nios2-linux-bugfix/src/scripts-trunk/gnu-ia-helper-mm -i /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 -l /scratch/sandra/nios2-linux-bugfix/logs -o /scratch/sandra/nios2-linux-bugfix/obj -p /scratch/sandra/nios2-linux-bugfix/pkg -s /scratch/sandra/nios2-linux-bugfix/src -T /scratch/sandra/nios2-linux-bugfix/testlogs -h i686-mingw32 -m iamm-jvm sgxx/2013.05/nios2-linux-gnu-lite
popenv

# task [215/219] /i686-mingw32/package_installanywhere/installer
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
pushenvvar CC 'i686-mingw32-gcc '
pushenvvar CXX 'i686-mingw32-g++ '
pushenvvar LD 'i686-mingw32-ld '
pushenvvar OBJCOPY 'i686-mingw32-objcopy '
pushenvvar AR 'i686-mingw32-ar '
pushenvvar RANLIB 'i686-mingw32-ranlib '
prepend_path PATH /scratch/sandra/nios2-linux-bugfix/obj/tools-i686-pc-linux-gnu-2013.05-43-nios2-linux-gnu-i686-mingw32/bin
/scratch/sandra/nios2-linux-bugfix/src/scripts-trunk/gnu-ia-helper -i /scratch/sandra/nios2-linux-bugfix/install/host-i686-mingw32 -l /scratch/sandra/nios2-linux-bugfix/logs -o /scratch/sandra/nios2-linux-bugfix/obj -p /scratch/sandra/nios2-linux-bugfix/pkg -s /scratch/sandra/nios2-linux-bugfix/src -T /scratch/sandra/nios2-linux-bugfix/testlogs -h i686-mingw32 -f /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras/getting_started/license.html -e /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras sgxx/2013.05/nios2-linux-gnu-lite
popenv

# task [216/219] /fini/extras_package
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.extras
pushd /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu
tar cjf /scratch/sandra/nios2-linux-bugfix/pkg/sourceryg++-2013.05-43-nios2-linux-gnu.extras.tar.bz2 --owner=0 --group=0 sourceryg++-2013.05-43-nios2-linux-gnu.extras
popd
popenv

# task [217/219] /fini/build_summary
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
cat > /scratch/sandra/nios2-linux-bugfix/pkg/sourceryg++-2013.05-43-nios2-linux-gnu-summary.xml <<'EOF0'
<?xml version="1.0" ?><build><variable var="release_config_canonical" value="sgxx/2013.05/nios2-linux-gnu-lite"/><variable var="version" value="2013.05-43"/><variable var="hosts_string" value="i686-pc-linux-gnu i686-mingw32"/><variable var="target" value="nios2-linux-gnu"/><variable var="build_date" value="20130625"/><variable var="HOSTNAME" value="build2-lucid-cs"/><variable var="build_operating_system" value="squeeze/sid"/><variable var="build_uname" value="Linux build2-lucid-cs 2.6.32-45-generic #104-Ubuntu SMP Tue Feb 19 21:20:09 UTC 2013 x86_64 GNU/Linux"/><variable var="USER" value="sandra"/><checksum type="md5" value="ae18e084ab602a14aed4b1c2f46bc701" file="sourceryg++-2013.05-43-nios2-linux-gnu-i686-pc-linux-gnu.tar.bz2"/><checksum type="md5" value="3ea7e8a9bfdea822ff4769ab4a649158" file="sourceryg++-2013.05-43-nios2-linux-gnu-i686-mingw32.tar.bz2"/><checksum type="md5" value="fb1998dc0aa05b16052a01b511f986ea" file="sourceryg++-2013.05-43-nios2-linux-gnu.extras.tar.bz2"/></build>
EOF0
popenv

# task [218/219] /fini/backups_package
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu.backup
pushd /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu
tar cjf /scratch/sandra/nios2-linux-bugfix/pkg/sourceryg++-2013.05-43-nios2-linux-gnu.backup.tar.bz2 --owner=0 --group=0 sourceryg++-2013.05-43-nios2-linux-gnu.backup
popd
popenv

# task [219/219] /fini/sources_package
pushenv
pushenvvar CC_FOR_BUILD i686-pc-linux-gnu-gcc
pushenvvar CXX_FOR_BUILD i686-pc-linux-gnu-g++
mkdir -p /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
cp /scratch/sandra/nios2-linux-bugfix/pkg/sourceryg++-2013.05-43-nios2-linux-gnu-summary.xml /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
cp /scratch/sandra/nios2-linux-bugfix/logs/sourceryg++-2013.05-43-nios2-linux-gnu.sh /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu/sourceryg++-2013.05-43-nios2-linux-gnu
pushd /scratch/sandra/nios2-linux-bugfix/obj/pkg-2013.05-43-nios2-linux-gnu
tar cjf /scratch/sandra/nios2-linux-bugfix/pkg/sourceryg++-2013.05-43-nios2-linux-gnu.src.tar.bz2 --owner=0 --group=0 sourceryg++-2013.05-43-nios2-linux-gnu
popd
/scratch/sandra/nios2-linux-bugfix/src/scripts-trunk/gnu-test -i /scratch/sandra/nios2-linux-bugfix/install -l /scratch/sandra/nios2-linux-bugfix/logs -o /scratch/sandra/nios2-linux-bugfix/obj -p /scratch/sandra/nios2-linux-bugfix/pkg -s /scratch/sandra/nios2-linux-bugfix/src -T /scratch/sandra/nios2-linux-bugfix/testlogs -T /scratch/sandra/nios2-linux-bugfix/obj/testlogs-2013.05-43-nios2-linux-gnu sgxx/2013.05/nios2-linux-gnu-lite
copy_dir /scratch/sandra/nios2-linux-bugfix/obj/testlogs-2013.05-43-nios2-linux-gnu /scratch/sandra/nios2-linux-bugfix/testlogs
/scratch/sandra/nios2-linux-bugfix/src/scripts-trunk/gnu-test-package -i /scratch/sandra/nios2-linux-bugfix/install -l /scratch/sandra/nios2-linux-bugfix/logs -o /scratch/sandra/nios2-linux-bugfix/obj -p /scratch/sandra/nios2-linux-bugfix/pkg -s /scratch/sandra/nios2-linux-bugfix/src -T /scratch/sandra/nios2-linux-bugfix/testlogs -T /scratch/sandra/nios2-linux-bugfix/obj/testlogs-2013.05-43-nios2-linux-gnu sgxx/2013.05/nios2-linux-gnu-lite
